# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Defines enumeration classes for flags, with which certain behaviors of the
connectors can be configured.
"""

import enum

__all__ = ("Laziness", "Parallelization")


@enum.unique
class Laziness(enum.IntEnum):
    """An enumeration type for defining the laziness of input connectors. The
    enumeration values are sorted by how lazy the behavior is, which they represent,
    so smaller/greater comparisons are possible. Do not rely on the exact integer
    value though, since they are not guaranteed to have the same value in any version
    of this package.

    1. ON_REQUEST
        the setter is called, when its execution is requested by a method further
        down the processing chain.
    2. ON_NOTIFY
        the setter is called, when the connected getter has computed the input
        value, even if the execution of this setter has not been requested.
    3. ON_ANNOUNCE
        the setter is requests its input value immediately and is executed as soon
        as that is available. But connecting the setter, does not cause a request
        of the input value.
    4. ON_CONNECT
        same as ON_ANNOUNCE, but the value is also requested, when a new connection
        is established, which influences the input value of this connector. This
        connection is does not necessarily have to be with this connector, but it
        can also be further upstream in the processing chain.
    """
    ON_REQUEST = 1
    ON_NOTIFY = 2
    ON_ANNOUNCE = 3
    ON_CONNECT = 4

    @staticmethod
    def get_default():
        """returns the default laziness setting for input connectors.

        :returns: a constant from this enumeration
        """
        return Laziness.ON_REQUEST


@enum.unique
class Parallelization(enum.Enum):
    """An enumeration type for the parallelization parameter of an executor's :meth:`~connectors._common._executors.Executor.run_method` method:

    * SEQUENTIAL
        the method can only be executed sequentially.
    * THREAD
        the method should be executed in a separate thread, sequential execution
        is possible as a fallback.
    * PROCESS
        the method should be executed in a separate process, threaded execution
        or sequential execution can be used as a fallback.
    """
    SEQUENTIAL = enum.auto()
    THREAD = enum.auto()
    PROCESS = enum.auto()

    @staticmethod
    def get_default_output_parallelization():
        """returns the default parallelization setting for output connectors.

        :returns: a constant from this enumeration
        """
        return Parallelization.THREAD

    @staticmethod
    def get_default_input_parallelization():
        """returns the default parallelization setting for input connectors.

        :returns: a constant from this enumeration
        """
        return Parallelization.SEQUENTIAL

    @staticmethod
    def get_default_multiinput_parallelization():
        """returns the default parallelization setting for multi-input connectors.

        :returns: a constant from this enumeration
        """
        return Parallelization.SEQUENTIAL



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Base classes for the connector classes"""

import functools
import weakref

__all__ = ("Connector", "InputConnector")


class Connector:
    """Base class for connectors.
    Connectors are objects that replace methods of a class so that they can be
    connected to each other. This way changing data at one end of a connection
    chain will automatically cause the data at the other end of the chain to be
    updated, when that data is retrieved the next time.
    """

    def __init__(self, instance, method, parallelization, executor):
        """
        :param instance: the instance of which the method is replaced by this connector
        :param method: the unbound method that is replaced by this connector
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :meth:`set_parallelization` method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :meth:`~connectors.connectors.Connector.set_executor`
                         method for details
        """
        self._instance = weakref.ref(instance)  # the weak reference avoids reference counting errors due to circular references
        self._method = method
        self._parallelization = parallelization
        self._executor = executor
        functools.update_wrapper(self, method)

    def __call__(self, *args, **kwargs):
        """By making the object callable, it mimics the replaced method.
        This is a virtual method which shall be overridden so it calls the
        replaced method in the expected manner.

        :param `*args,**kwargs`: arguments for the replaced method
        :returns: the return value of the replaced method
        """
        raise NotImplementedError("This method should have been implemented in a derived class")

    def connect(self, connector):
        """Abstract method that defines the interface of a :class:`~connectors.connectors.Connector`
        for connecting it with other connectors.

        :param connector: the :class:`~connectors.connectors.Connector` instance
                          to which this connector shall be connected
        :returns: the instance of which this :class:`~connectors.connectors.Connector`
                  has replaced a method
        """
        raise NotImplementedError("This method should have been implemented in a derived class")

    def disconnect(self, connector):
        """Abstract method that defines the interface of a :class:`~connectors.connectors.Connector`
        for disconnecting it from a connector, to which it is currently connected.

        :param connector: a :class:`~connectors.connectors.Connector` instance
                          from which this connector shall be disconnected
        :returns: the instance of which this :class:`~connectors.connectors.Connector`
                  has replaced a method
        """
        raise NotImplementedError("This method should have been implemented in a derived class")

    def set_parallelization(self, parallelization):
        """Specifies, if and how the execution of this connector can be parallelized.
        The choices are no parallelization, the execution in a separate thread
        and the execution in a separate process.
        This method specifies a hint, which level of parallelization is possible
        with the connector. If the executor of the connector, through which the
        computation is started, does not support the specified level, the next simpler
        one will be chosen. E.g. if a connector can be parallelized in a separate
        process, but the executor only allows threads or sequential execution, the
        connector will be executed in a separate thread.

        :param parallelization: a flag from the :class:`connectors.Parallelization` enum
        """
        self._parallelization = parallelization

    def set_executor(self, executor):
        """Sets the executor, which handles the computations, when the data is
        retrieved through this connector.
        An executor can be created with the :func:`connectors.executor` function. It
        manages the order and the parallelization of the computations, when updating
        the data in a processing chain.
        If multiple connectors in a processing chain need to be computed, the
        executor of the connector, which started the computations, is used for
        all computations.

        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function
        """
        self._executor = executor

    def _announce(self, connector, non_lazy_inputs):
        """This method is to notify other connectors in a processing chain,
        when a value is about to change.
        This way, this connector knows, that it has to retrieve new data from the
        given connector and recompute its own value, when that value is requested.

        :param connector: a connector on whose value change this connector's value depends
        :param non_lazy_inputs: a :class:`~connectors._common._non_lazy_inputs.NonLazyInputs`
                                instance to which input connectors can be appended,
                                if they request an immediate re-computation (see
                                the :class:`~connectors._connectors._baseclasses.InputConnector`'s
                                :meth:`~connectors._connectors._baseclasses.InputConnector.set_laziness`
                                method for more about lazy execution)
        """
        raise NotImplementedError("This method should have been implemented in a derived class")

    def _get_instance(self):
        """A method, that is used internally by the *Connectors* package to
        retrieve the object instance of which the connector has replaced a method.
        Even though the connectors store weak references to the instance, the return
        value of this method is a strong reference.

        :returns: a Python object
        """
        return self._instance()

    def _get_connector(self):
        """A method, that is used internally by the *Connectors* package for
        compatibility with :class:`~connectors._proxies._baseclasses.ConnectorProxy`
        instances.
        In a :class:`~connectors._proxies._baseclasses.ConnectorProxy`, this method
        creates the connector for the corresponding method, replaces that method
        with it and returns the connector. With this method, a Connector mimics
        this behavior, so it is not necessary to check, whether a method has
        already been replaced by its connector, or if it is still managed by a
        :class:`~connectors._proxies._baseclasses.ConnectorProxy`.

        :returns: ``self``
        """
        return self


class InputConnector(Connector):
    """Base class for input connectors, that replace setter methods."""
    # pylint: disable=abstract-method; pylint shall not complain, that the __call__ and _announce-methods are not overridden in InputConnector

    def __init__(self, instance, method, laziness, parallelization, executor):
        """
        :param instance: the instance of which the method is replaced by this connector
        :param method: the unbound method that is replaced by this connector
        :param laziness: a flag from the :class:`connectors.Laziness` enum. See
                         the :meth:`set_laziness` method for details
        :param parallelization: a flag from the :class:`~connectors.Parallelization` enum.
                                See the :meth:`~connectors._connectors._baseclasses.InputConnector.set_parallelization`
                                method for details
        :param executor: an :class:`~connectors._common._executors.Executor`
                         instance, that can be created with the :func:`connectors.executor`
                         function. See the :meth:`~connectors._connectors._baseclasses.InputConnector.set_executor`
                         method for details
        """
        Connector.__init__(self, instance, method, parallelization, executor)
        self._laziness = laziness

    def connect(self, connector):
        """Connects this :class:`~connectors._connectors._baseclasses.InputConnector` to an output.

        :param connector: the :class:`~connectors.connectors.Connector` instance to which this connector shall be connected
        :returns: the instance of which this :class:`~connectors._connectors._baseclasses.InputConnector`
                  has replaced a method
        """
        connector.connect(self)
        return self._instance()

    def disconnect(self, connector):
        """Disconnects this :class:`~connectors._connectors._baseclasses.InputConnector`
        from an output, to which is has been connected.

        :param connector: a :class:`~connectors.connectors.Connector` instance from which this connector shall be disconnected
        :returns: the instance of which this :class:`~connectors.connectors.Connector` has replaced a method
        """
        connector.disconnect(self)
        return self._instance()

    def set_laziness(self, laziness):
        """Configures the lazy execution of the connector.
        Normally the connectors are executed lazily, which means, that any computation
        is only started, when the result of a processing chain is requested. For
        certain use cases it is necessary to disable this lazy execution, though,
        so that the values are updated immediately as soon as new data is available.
        There are different behaviors for the (non) lazy execution, which are
        described in the :class:`connectors.Laziness` enum.

        :param laziness: a flag from the :class:`connectors.Laziness` enum
        """
        self._laziness = laziness



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Base classes for the connector decorators"""

from connectors import _common as lib

__all__ = ("ConnectorDecorator", "InputDecorator")

default_executor = lib.executor()   # this instance is for the default value of the connector decorator's constructors


class ConnectorDecorator:
    """An abstract base class for the decoration of methods so they can be used
    as connectors.
    """

    def __init__(self, parallelization, executor):
        """
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :class:`~connectors.connectors.Connector`'s
                                :meth:`~connectors.connectors.Connector.set_parallelization` method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :class:`~connectors.connectors.Connector`'s
                         :meth:`~connectors.connectors.Connector.set_executor` method
                         for details
        """
        self._method = None     # Will be set in __call__
        self._parallelization = parallelization
        self._executor = executor

    def __call__(self, method):
        """Is called in order to replace the decorated method with this decorator.

        :param method: the unbound method, that was decorated with this decorator
        :returns: this decorator
        """
        self._method = method
        return self

    def __get__(self, instance, instance_type):
        """Is called, when the decorated method is accessed.

        :param instance: the instance of which a method shall be replaced
        :param instance_type: the type of the instance
        :returns: a :class:`connectors._proxies._baseclasses.ConnectorProxy` instance,
                  that mimics the decorated method and adds the connector functionality
        """
        raise NotImplementedError("This method should have been overridden in a derived class")


class InputDecorator(ConnectorDecorator):
    """An addition to a :class:`ConnectorDecorator` for input connectors that notify
    their instances' output connectors when an incoming value changes their result
    values.
    """
    # pylint: disable=abstract-method; pylint shall not complain, that the __get__-method is not overridden in InputDecorator

    def __init__(self,
                 observers=(),
                 laziness=lib.Laziness.get_default(),
                 parallelization=lib.Parallelization.get_default_input_parallelization(),
                 executor=default_executor):
        """
        :param observers: the names of output methods that are affected by passing
                          a value to this connector. For convenience it is also
                          possible to pass a string here, if only one output
                          connector depends on this input.
        :param laziness: a flag from the :class:`connectors.Laziness` enum. See the
                         :class:`~connectors.connectors.SingleInputConnector`'s
                         :meth:`~connectors.connectors.SingleInputConnector.set_laziness`
                         method for details
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :class:`~connectors.connectors.MultiInputConnector`'s
                                :meth:`~connectors.connectors.SingleInputConnector.set_parallelization`
                                method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :class:`~connectors._connectors._baseclasses.InputConnector`'s
                         :meth:`~connectors._connectors._baseclasses.InputConnector.set_executor` method
                         for details
        """
        ConnectorDecorator.__init__(self, parallelization=parallelization, executor=executor)
        if isinstance(observers, str):
            self._observers = (observers,)
        else:
            self._observers = observers
        self._laziness = laziness
        self._announce_condition = None
        self._notify_condition = None

    def announce_condition(self, method):
        """A decorator, that can be used as a method of the connector method, to
        define a condition for the propagation of announcements through the
        input connector.

        The decorated method shall return ``True``, if the announcement shall be
        propagated and ``False``` otherwise. For normal input connectors, it shall
        not require any arguments, while for multi-input connectors, it must accept
        the data ID of the connection, through which the announcement was received.

        Before the values in a processing chain are updated, the value changes
        are announced to the downstream processors. Only if data is retrieved
        through an output connector, that has pending announcements, the actual
        value changes are requested from the upstream processors (lazy execution).
        If an input connector has defined a condition on the propagation of
        announcements and this condition evaluates to ``False``, the announcements
        are not forwarded to the downstream processors. This also prevents those
        processors from requesting updated values from upstream.

        The usage is described by the following example: if `A` is the name of
        the method, that is decorated with :class:`connectors.Input`
        or :class:`connectors.MultiInput`, the method for the condition has to be
        decorated with ``@A.announce_condition``.

        :param: the method, that defines the condition
        :returns: the same method
        """
        self._announce_condition = method
        return method

    def notify_condition(self, method):
        """A decorator, that can be used as a method of the connector method, to
        define a condition for notifying the observing output connectors about
        a value, that has been changed by this connector.

        The decorated method shall return ``True``, if the notification shall be
        sent and ``False`` otherwise. For normal input connectors, it shall accept
        the new value as an argument, while for multi-input connectors, it must
        accept the data ID of the connection, through which the announcement was
        received and the new value.

        This condition is checked after the input connector (the setter method)
        has been executed. If an input connector has defined a condition on the
        notification of its observing output connectors and this condition evaluates
        to ``False``, the output connectors are sent a cancel notification, that
        informs them, that the state of the object, to which the connectors belong,
        has not changed in a relevant way. This prevents the update of values
        further down the processing chain.

        The usage is described by the following example: if `A` is the name of
        the method, that is decorated with :class:`connectors.Input`
        or :class:`connectors.MultiInput`, the method for the condition has to be
        decorated with ``@A.notify_condition``.

        :param: the method, that defines the condition
        :returns: the same method
        """
        self._notify_condition = method
        return method



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains the decorator and connector classes for the macro input connector."""

import functools
from ._baseclass import MacroDecorator

__all__ = ("MacroInput",)


class MacroInput(MacroDecorator):
    """A decorator to replace a method with a macro input connector.

    Macro connectors are useful, when a processing network shall be encapsulated
    in a class. In such a case, macro input connectors are used to export input
    connectors from the internal processing network to be accessible as connectors
    of the encapsulating class's instance.

    The decorated method must not take any parameters and yield all input connectors
    from the internal processing network, that it exports. Exporting multiple
    connectors through the same macro connector is possible and useful, when all
    of these connectors shall always receive the same input value.

    The resulting connector will behave like an input connector, which is very
    different from the decorated method:

    - the connector's arguments are passed to all the exported input connectors,
      when it is called.
    - when called, the connector returns the instance to which it belongs, so
      that changing a parameter and retrieving a result in one line is possible
    - when a behavior (e.g. laziness) of the connector is changed, the change
      is passed on to all the exported connectors.
    """

    def __get__(self, instance, instance_type):
        """Is called, when the decorated method is accessed.

        :param instance: the instance of which a method shall be replaced
        :param instance_type: the type of the instance
        :returns: a :class:`MacroInputConnector` instance
        """
        return MacroInputConnector(instance=instance, method=self._method)


class MacroInputConnector:
    """A Connector-class that exports input connectors from an internal processing
    network to the API of the class, that encapsulates the network.
    """

    def __init__(self, instance, method):
        """
        :param instance: the instance in which the method is replaced by this connector
        :param method: the unbound method, that is replaced by this connector
        """
        self.__instance = instance
        self.__method = method
        functools.update_wrapper(self, method)

    def __call__(self, *args, **kwargs):
        """Calls all input connectors, that are exported by this, with the given
        parameters.

        :param `*args,**kwargs`: parameters with which the exported input connectors shall be called
        :returns: the instance of which this connector has replaced a method
        """
        for connector in self.__method(self.__instance):
            connector(*args, **kwargs)
        return self.__instance

    def set_laziness(self, laziness):
        """Configures the lazy execution of the connector.
        Normally the connectors are executed lazily, which means, that any computation
        is only started, when the result of a processing chain is requested. For
        certain use cases it is necessary to disable this lazy execution, though,
        so that the values are updated immediately as soon as new data is available.
        There are different behaviors for the (non) lazy execution, which are
        described in the :class:`connectors.Laziness` enum.

        :param laziness: a flag from the :class:`connectors.Laziness` enum
        """
        for connector in self.__method(self.__instance):
            connector.set_laziness(laziness)

    def set_parallelization(self, parallelization):
        """Specifies, if and how the execution of this connector can be parallelized.
        The choices are no parallelization, the execution in a separate thread
        and the execution in a separate process.
        This method specifies a hint, which level of parallelization is possible
        with the connector. If the executor of the connector, through which the
        computation is started, does not support the specified level, the next simpler
        one will be chosen. E.g. if a connector can be parallelized in a separate
        process, but the executor only allows threads or sequential execution, the
        connector will be executed in a separate thread.

        :param parallelization: a flag from the :class:`connectors.Parallelization` enum
        """
        for connector in self.__method(self.__instance):
            connector.set_parallelization(parallelization)

    def set_executor(self, executor):
        """Sets the executor, which handles the computations, when the data is
        retrieved through this connector.
        An executor can be created with the :func:`connectors.executor` function. It
        manages the order and the parallelization of the computations, when updating
        the data in a processing chain.
        If multiple connectors in a processing chain need to be computed, the
        executor of the connector, which started the computations, is used for
        all computations.

        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function
        """
        for connector in self.__method(self.__instance):
            connector.set_executor(executor)

    def connect(self, connector):
        """Connects all exported input connectors to the given output.

        :param connector: the output connector to which the exported connectors shall be connected
        :returns: the instance of which this connector has replaced a method
        """
        connector.connect(self)
        return self.__instance

    def disconnect(self, connector):
        """Disconnects all exported input connectors from the given output.

        :param connector: the output connector from which the exported connectors shall be disconnected
        :returns: the instance of which this connector has replaced a method
        """
        connector.disconnect(self)
        return self.__instance

    def _connect(self, connector):
        """This method is called from an :class:`OutputConnector`, when it is  being
        connected to this :class:`MacroInputConnector`.

        :param connector: the :class:`OutputConnector` instance to which this connector shall be connected
        :returns: yields all the :class:`InputConnector`s that this exports
        """
        for c1 in self.__method(self.__instance):
            for c2 in c1._connect(connector):
                yield c2

    def _disconnect(self, connector):
        """This method is called from an :class:`OutputConnector`, when it is  being
        disconnected from this :class:`MacroInputConnector`.

        :param connector: the :class:`OutputConnector` instance from which this connector shall be disconnected
        :returns: yields all the :class:`InputConnector`s that this exports
        """
        for c1 in self.__method(self.__instance):
            for c2 in c1._disconnect(connector):
                yield c2



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains a derived class of asyncio.Event"""

import asyncio

__all__ = ("Event",)


class Event(asyncio.Event):
    """Extends asyncio.Event, so that the event loop for waiting for the Event
    is passed to the wait method.

    This makes it possible to create instances of this event before starting the
    event loop.
    """

    async def wait(self, executor):     # pylint: disable=arguments-differ; the whole point of this class is, that the executor is passed to the wait method
        """waits for the Event's value to be "True"

        :param executor: the executor instance, that manages the event loop
        :returns: True
        """
        self._loop = executor.get_event_loop()
        try:
            return await super().wait()
        finally:
            self._loop = None



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains processing blocks for common tasks with the *Connectors* package."""

from ._multiplexer import *
from ._passthrough import *
from ._weakrefproxygenerator import *



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains a base class for the macro connector decorators."""


class MacroDecorator:
    """Base class for the decorators for macro connectors"""

    def __init__(self):
        self._method = None

    def __call__(self, method):
        """Is called in order to replace the decorated method with this decorator.

        :param method: the unbound method, that was decorated with this decorator
        :returns: this decorator
        """
        self._method = method
        return self

    def __get__(self, instance, instance_type):
        """Is called, when the decorated method is accessed.

        :param instance: the instance of which a method shall be replaced
        :param instance_type: the type of the instance
        :returns: a :class:`MacroInputConnector` or :class:`MacroOutputConnector` instance
        """
        raise NotImplementedError("This method should have been implemented in a derived class")



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains classes for the multi-input proxies"""

from ..connectors import Connector, MultiInputConnector, ConditionalMultiInputConnector
from .. import _common as common
from ._input import SingleInputProxy

__all__ = ("MultiInputProxy",)


class ReplaceMethod:
    """Mimics a replace method by removing the old value and adding the new one.
    Instances of this class will be created, when no replace method is specified
    for a multi-input connector.
    """

    def __init__(self, add_method, remove_method):
        """
        :param method: the unbound method, that is replaced by the multi-input connector
        :param remove_method: the unbound method, that is used to remove data, that
                              has been added through the multi-input connector
        """
        self.__add = add_method
        self.__remove = remove_method

    def __call__(self, instance, data_id, value):
        """
        :param instance: the instance in which the method has been replaced by the multi-input connector
        :param data_id: the ID under which the data, that shall be replaced, has been stored
        :param value: the new value
        :returns: the new data ID, under which the new data is stored
        """
        self.__remove(instance, data_id)
        return self.__add(instance, value)


class MultiInputProxy(SingleInputProxy):
    """A proxy class for multi-input connectors.
    Connector proxies are returned by the connector decorators, while the methods
    are replaced by the actual connectors. Think of a connector proxy like of a
    bound method, which is also created freshly, whenever a method is accessed.
    The actual connector only has a weak reference to its instance, while this
    proxy has a hard reference, but mimics the connector in almost every other
    way. This makes constructions like ``value = Class().connector()`` possible.
    Without the proxy, the instance of the class would be deleted before the connector
    method is called, so that the weak reference of the connector would be expired
    during its call.
    """

    def __init__(self, instance, method,                            # pylint: disable=too-many-arguments; this constructor may be complicated, since the class is instantiated through the decorators, which have a much simpler API
                 remove_method, replace_method,
                 observers, announce_condition, notify_condition,
                 laziness, parallelization, executor):
        """
        :param instance: the instance in which the method is replaced by this connector proxy
        :param method: the unbound method, that is replaced by this connector proxy
        :param remove_method: an unbound method, that is used to remove data, that
                              has been added through this connector proxy
        :param replace_method: an unbound method, that is used to replace data,
                               that has been added through this connector proxy
        :param observers: the names of output methods that are affected by passing
                          a value to this connector proxy
        :param announce_condition: a method, that defines the condition for the
                                   announcements to the observing output connectors.
                                   This method must accept the data ID of the changed
                                   output connector as an argument in addition to
                                   ``self``
        :param notify_condition: a method, that defines the condition for the
                                 notifications to the observing output connectors.
                                 This method must accept the data ID and the new
                                 input value as an argument in addition to ``self``
        :param laziness: a flag from the :class:`connectors.Laziness` enum. See the
                         :meth:`set_laziness` method for details
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :meth:`set_parallelization` method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :class:`~connectors.connectors.MultiInputConnector`'s
                         :meth:`~connectors.connectors.MultiInputConnector.set_executor`
                         method for details
        """
        SingleInputProxy.__init__(self,
                                  instance=instance,
                                  method=method,
                                  observers=observers,
                                  announce_condition=announce_condition,
                                  notify_condition=notify_condition,
                                  laziness=laziness,
                                  parallelization=parallelization,
                                  executor=executor)
        self.__remove = remove_method
        if replace_method is None:
            self.__replace = ReplaceMethod(add_method=method, remove_method=remove_method)
        else:
            self.__replace = replace_method

    def __getitem__(self, key):
        """Allows to use a multi-input connector as multiple single-input connectors.
        The key, under which a virtual single-input connector is accessed, shall
        also be returned the data ID, under which the result of the connected
        output is stored.

        :param key: a key for accessing a particular virtual single-input connector
        :returns: a MultiinputItem, which enhances the decorated method with the
                  functionality of the virtual single-input connector
        """
        for o in self._observers:
            if isinstance(getattr(self._get_instance(), o), Connector):
                return self._get_connector()[key]
        return common.MultiinputItem(connector=self,
                                     instance=self._get_instance(),
                                     replace_method=self.__replace,
                                     key=key,
                                     observers=(),
                                     executor=self._executor)

    def _create_connector(self, instance, method, parallelization, executor):
        """Creates and returns the multi-input connector.

        :param instance: the instance in which the method is replaced by the connector
        :param method: the unbound method that is replaced by the connector
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :meth:`set_parallelization` method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :class:`~connectors.connectors.MultiInputConnector`'s
                         :meth:`~connectors.connectors.MultiInputConnector.set_executor`
                         method for details
        :returns: an :class:`MultiInputConnector` instance
        """
        if self._announce_condition is None and self._notify_condition is None:
            return MultiInputConnector(instance=instance,
                                       method=method,
                                       remove_method=self.__remove,
                                       replace_method=self.__replace,
                                       observers=self._observers,
                                       laziness=self._laziness,
                                       parallelization=parallelization,
                                       executor=executor)
        else:
            announce_condition, notify_condition = common.select_condition_methods(self._announce_condition,
                                                                                   self._notify_condition)
            return ConditionalMultiInputConnector(instance=instance,
                                                  method=method,
                                                  remove_method=self.__remove,
                                                  replace_method=self.__replace,
                                                  observers=self._observers,
                                                  announce_condition=announce_condition,
                                                  notify_condition=notify_condition,
                                                  laziness=self._laziness,
                                                  parallelization=parallelization,
                                                  executor=executor)



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains :class:`ConnectorProxy`, the base class for connector proxies."""

import functools


class ConnectorProxy:
    """A base class for proxy objects of connectors.
    Connector proxies are returned by the connector decorators, while the methods
    are replaced by the actual connectors. Think of a connector proxy like of a
    bound method, which is also created freshly, whenever a method is accessed.
    The actual connector only has a weak reference to its instance, while this
    proxy has a hard reference, but mimics the connector in almost every other
    way. This makes constructions like ``value = Class().connector()`` possible.
    Without the proxy, the instance of the class would be deleted before the connector
    method is called, so that the weak reference of the connector would be expired
    during its call.
    """

    def __init__(self, instance, method, parallelization, executor):
        """
        :param instance: the instance in which the method is replaced by this connector proxy
        :param method: the unbound method that is replaced by this connector proxy
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :meth:`set_parallelization` method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :meth:`~connectors._proxies._baseclasses.ConnectorProxy.set_executor`
                         method for details
        """
        self.__instance = instance
        self.__method = method
        self._parallelization = parallelization
        self._executor = executor
        self.__connector = None
        functools.update_wrapper(self, method)

    def __call__(self, *args, **kwargs):
        """By making the object callable, it mimics the replaced method.

        :param `*args,**kwargs`: possible arguments for the replaced method
        """
        return self.__method(self.__instance, *args, **kwargs)

    def connect(self, connector):
        """Connects this connector with another one.

        :param connector: the :class:`~connectors.connectors.Connector` instance
                          to which this connector shall be connected
        :returns: the instance of which this :class:`~connectors.connectors.Connector`
                  has replaced a method
        """
        return self._get_connector().connect(connector)

    def set_parallelization(self, parallelization):
        """Specifies, if and how the execution of this connector can be parallelized.
        The choices are no parallelization, the execution in a separate thread
        and the execution in a separate process.
        This method specifies a hint, which level of parallelization is possible
        with the connector. If the executor of the connector, through which the
        computation is started, does not support the specified level, the next simpler
        one will be chosen. E.g. if a connector can be parallelized in a separate
        process, but the executor only allows threads or sequential execution, the
        connector will be executed in a separate thread.

        :param parallelization: a flag from the :class:`connectors.Parallelization` enum
        """
        self._get_connector().set_parallelization(parallelization)

    def set_executor(self, executor):
        """Sets the executor, which handles the computations, when the data is
        retrieved through this connector.
        An executor can be created with the :func:`connectors.executor` function. It
        manages the order and the parallelization of the computations, when updating
        the data in a processing chain.
        If multiple connectors in a processing chain need to be computed, the
        executor of the connector, which started the computations, is used for
        all computations.

        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function
        """
        self._get_connector().set_executor(executor)

    def _get_instance(self):
        """A method, that is used internally by the *Connectors* package to retrieve
         the object instance of which the connector proxy has replaced a method.

        :returns: a Python object
        """
        return self.__instance

    def _get_connector(self):
        """Creates the actual connector, that is represented by this proxy, replaces
        the method with it and returns the connector. This proxy is no longer used
        after this.

        :returns: a Connector instance
        """
        self.__replace_method()
        return self.__connector

    def _create_connector(self, instance, method, parallelization, executor):
        """Abstract method, in which derived classes can implement the creation
        of their connectors.

        :param instance: the instance in which the method is replaced by the connector
        :param method: the unbound method that is replaced by the connector
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :meth:`set_parallelization` method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :meth:`~connectors._proxies._baseclasses.ConnectorProxy.set_executor`
                         method for details
        :returns: a :class:`~connectors.connectors.Connector` instance
        """
        raise NotImplementedError("This method should have been implemented in a derived class")

    def __replace_method(self):
        """Replaces the method with the connector
        This proxy is no longer used after this.
        """
        if self.__connector is None:
            self.__connector = self._create_connector(instance=self.__instance,
                                                      method=self.__method,
                                                      parallelization=self._parallelization,
                                                      executor=self._executor)
            setattr(self.__instance, self.__method.__name__, self.__connector)



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains the :class:`WeakrefProxyGenerator` class."""

import weakref
import connectors

__all__ = ("WeakrefProxyGenerator",)


class WeakrefProxyGenerator:
    """A helper class to reduce memory usage in certain processing chains by
    discarding intermediate results. It takes an object as input and outputs a
    weak reference proxy to it.
    The deletion of the strong reference to the input object can triggered by connecting
    the output connector for the final result to this class's :meth:`~WeakrefProxyGenerator.delete_reference`
    method. This will cause the input object to be garbage collected, if no further
    references to it exist.
    This class should be used in combination with the caching of the final result,
    since this prevents, that the deleted input data is required to re-compute the
    result, when it's retrieved repeatedly.
    """

    def __init__(self, data=None):
        """
        :param data: the input object
        """
        self.__data = data

    @connectors.Output(caching=False)
    def output(self):
        """
        Creates and returns a weak reference proxy to the input object.
        As long as the object has not been garbage collected, this proxy should
        behave exactly as the input object.

        :returns: a weak reference proxy to the input object, if it can be weakly
                  referenced, otherwise the object itself
        """
        try:
            return weakref.proxy(self.__data)
        except TypeError:       # if the data does not allow weak references to it
            return self.__data

    @connectors.Input("output")
    def input(self, data):
        """Specifies the input object.

        :param data: the input object
        :returns: the WeakrefProxyGenerator instance
        """
        self.__data = data
        return self

    @connectors.Input(laziness=connectors.Laziness.ON_NOTIFY)
    def delete_reference(self, *args, **kwargs):    # pylint: disable=unused-argument; passing arguments to the delete_reference method shall not raise an error, so it can be connected to any output
        """Causes the strong reference to the input object to be deleted, so that
        the input object can be garbage collected.
        This connector should be connected to the output of the final result, so
        it is notified, when the result has been computed and the input object is
        no longer required.

        :param `*args,**kwargs`: these parameters just there fore compatibility with
                                 other input connectors and are not used in this method
        :returns: the WeakrefProxyGenerator instance
        """
        self.__data = None
        return self



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains the MultiInputConnector class"""

import asyncio
import collections
import weakref
from .. import _common as common
from ._baseclasses import InputConnector

__all__ = ("MultiInputConnector", "ConditionalMultiInputConnector")


class MultiInputConnector(InputConnector):
    """A Connector-class that replaces special setter methods, that allow to pass
    multiple values, so they can be used to connect different objects in a processing
    chain.
    """

    def __init__(self, instance, method, remove_method, replace_method, observers, laziness, parallelization, executor):
        """
        :param instance: the instance of which the method is replaced by this connector
        :param method: the unbound method, that is replaced by this connector
        :param remove_method: an unbound method, that is used to remove data, that
                              has been added through this connector
        :param replace_method: an unbound method, that is used to replace data,
                               that has been added through this connector
        :param observers: the names of output methods that are affected by passing a value to this connector
        :param laziness: a flag from the :class:`connectors.Laziness` enum. See
                         the :meth:`~connectors.connectors.MultiInputConnector.set_laziness`
                         method for details
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :meth:`~connectors.connectors.MultiInputConnector.set_parallelization`
                                method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :meth:`~connectors.connectors.MultiInputConnector.set_executor`
                         method for details
        """
        InputConnector.__init__(self, instance, method, laziness, parallelization, executor)
        self.__remove = remove_method
        self.__replace = replace_method
        self.__observers = common.resolve_observers(instance=instance, observers=observers)
        self._connections = weakref.WeakKeyDictionary()
        self.__announcements = set()
        self.__notifications = collections.OrderedDict()
        self.__running = False              # is used to prevent, that the setter is executed multiple times for the same changes
        self.__computable = common.Event()  # is set, when there is no pending announcement
        self.__computable.set()

    def __call__(self, *args, **kwargs):
        """By making the object callable, it mimics the replaced method.
        This method also notifies the output method that are affected by this call (observers).

        :param `*args,**kwargs`: parameters with which the replaced method has been called
        :returns: the return value of the replaced method
        """
        # announce the value change
        non_lazy_inputs = common.NonLazyInputs(common.Laziness.ON_ANNOUNCE)
        for o in self.__observers:
            o._announce(self, non_lazy_inputs)
        # call the replaced method
        self._executor.run_coroutine(self._request(self._executor))    # retrieve the announced values from the connectors first, so that everything is added in the correct order
        result = self._method(self._instance(), *args, **kwargs)
        # notify observers about the value change
        self._conditional_observer_notification(result, *args, **kwargs)
        # execute the non-lazy inputs
        non_lazy_inputs.execute(self._executor)
        # return the result of the method call
        return result

    def __getitem__(self, key):
        """Allows to use a multi-input connector as multiple single-input connectors.
        The key, under which a virtual single-input connector is accessed, shall
        also be returned the data ID, under which the result of the connected
        output is stored.

        :param key: a key for accessing a particular virtual single-input connector
        :returns: a MultiinputItem, which enhances the decorated method with the
                  functionality of the virtual single-input connector
        """
        return common.MultiinputItem(connector=self,
                                     instance=self._instance(),
                                     replace_method=self.__replace,
                                     key=key,
                                     observers=self.__observers,
                                     executor=self._executor)

    def _connect(self, connector, key=None):    # pylint: disable=arguments-differ; the key is only used by the multi-input connector
        """This method is called from an :class:`~connectors.OutputConnector`,
        when it is  being connected to this :class:`~connectors.MultiInputConnector`.

        :param connector: the output connector that shall be connected to this multi-input connector
        :param key: the data ID, under which the data of the given connector is stored
        :returns: yields self (see the :class:`~connectors.MacroInputConnector`,
                  where :meth:`~connectors.MacroInputConnector._connect` yields
                  all the :class:`~connectors.SingleInputConnector`s that it exports)
        """
        yield self
        self._connections[connector] = key
        self.__announcements.add(connector)
        self.__notifications[connector] = None  # create a placeholder in the notifications dict
        non_lazy_inputs = common.NonLazyInputs(situation=common.Laziness.ON_CONNECT)
        self._announce(connector, non_lazy_inputs=non_lazy_inputs)
        non_lazy_inputs.execute(self._executor)

    def _disconnect(self, connector):
        """This method is called from an :class:`~connectors.OutputConnector`,
        when it is  being disconnected from this :class:`~connectors.MultiInputConnector`.

        :param connector: the output connector from which this connector shall be disconnected
        :returns: yields self (see the :class:`~connectors.MacroInputConnector`,
                  where :meth:`~connectors.MacroInputConnector._disconnect` yields
                  all the :class:`~connectors.SingleInputConnector`s that it exports)
        """
        self.__announcements.discard(connector)
        if connector in self.__notifications:
            del self.__notifications[connector]
        non_lazy_inputs = common.NonLazyInputs(situation=common.Laziness.ON_CONNECT)
        common.forward_announcement(connector=self,
                                    observers=self.__observers,
                                    non_lazy_inputs=non_lazy_inputs,
                                    laziness=self._laziness)
        data_id = self._connections[connector]
        if data_id is not None:
            self.__remove(self._instance(), data_id)
        del self._connections[connector]
        self._conditional_observer_notification(data_id)
        non_lazy_inputs.execute(self._executor)
        yield self

    def _announce(self, connector, non_lazy_inputs):
        """This method is to notify this input connector, when a connected output
        connector can produce updated data.

        :param connector: the output connector whose value is about to change
        :param non_lazy_inputs: a :class:`~connectors._common._non_lazy_inputs.NonLazyInputs`
                                instance to which input connectors can be appended,
                                if they request an immediate re-computation (see
                                the :meth:`~connectors.connectors.MultiInputConnector.set_laziness`
                                method for more about lazy execution)
        """
        self.__computable.clear()
        self.__announcements.add(connector)
        common.forward_announcement(connector=self,
                                    observers=self.__observers,
                                    non_lazy_inputs=non_lazy_inputs,
                                    laziness=self._laziness)

    async def _notify(self, connector, value, executor):
        """This method is to notify this input connector, when a connected output
        connector has produced updated data.

        :param connector: the output connector whose value has changed
        :param value: the updated data from the output connector
        :param executor: the :class:`~connectors._common._executors.Executor`
                         instance, which managed the computation of the output
                         connector, and which shall be used for the computation
                         of this connector, in case it is not lazy.
        """
        self.__notifications[connector] = value
        self.__announcements.discard(connector)
        if not self.__announcements:
            self.__computable.set()
        if self._laziness == common.Laziness.ON_NOTIFY:
            await self._request(executor)

    def _cancel(self, connector):
        """Notifies this input connector, that an announced value change is not
        going to happen.

        :param connector: the output connector whose value change is canceled
        """
        self.__announcements.discard(connector)
        if not self.__announcements:
            self.__computable.set()
            for o in self.__observers:
                o._cancel(self)     # pylint: disable=protected-access; the _cancel method is meant to be called from other connectors, but not from outside this package

    async def _request(self, executor):
        """This method retrieves the updated data from the connected output connectors
        and recomputes this connector (if necessary).
        It is called by the output connectors, that are affected by this connector,
        (observers) when their values have to be computed.

        :param executor: the :class:`~connectors._common._executors.Executor` instance,
                         that manages the current computations
        """
        if not self.__running:
            self.__running = True
            try:
                # wait for the announced value changes
                if self.__announcements:
                    await asyncio.wait([a._request(executor) for a in self.__announcements])
                    await self.__computable.wait(executor)
                # execute the setter
                if self.__notifications:
                    for output_connector in self.__notifications:
                        data_id = self._connections[output_connector]
                        value = self.__notifications[output_connector]
                        if data_id is None:
                            self._connections[output_connector] = await executor.run_method(self._parallelization,
                                                                                            self._method,
                                                                                            self._instance(),
                                                                                            value)
                        else:
                            self._connections[output_connector] = await executor.run_method(self._parallelization,
                                                                                            self.__replace,
                                                                                            self._instance(),
                                                                                            data_id,
                                                                                            value)
                        self._conditional_observer_notification(data_id, value)
                        self._add_to_notification_condition_checks(data_id, value)
                    self.__notifications.clear()
                    if self._check_notification_condition():
                        for o in self.__observers:
                            o._notify(self)
            finally:
                self.__running = False

    def _conditional_observer_notification(self, data_id, *args, **kwargs):   # pylint: disable=unused-argument; this method has to be compatible with other input connectors
        """Notifies the observing output connectors, that this input has changed
        the instance's state.

        This method can be overridden by derived classes, for example to implement
        a conditional notification of the outputs.

        :param `*args,**kwargs`: the parameters, with which the replaced setter
                                method has been called (excluding ``self``)
        """
        for o in self.__observers:
            o._notify(self)

    def _add_to_notification_condition_checks(self, data_id, value):
        """A protected helper method to determine, if the observing output connectors
        have to be notified at the end of the :meth:`~connectors.MultiInputConnector._request`
        method.

        This method is meant to be overridden in derived classes, that implement
        a conditional notification of the observing output connectors.
        """

    def _check_notification_condition(self):   # pylint: disable=no-self-use; this method is a hook for derived classes
        """A protected helper method to determine, if the observing output connectors
        have to be notified at the end of the :meth:`~connectors.MultiInputConnector._request`
        method.

        This method is meant to be overridden in derived classes, that implement
        a conditional notification of the observing output connectors.
        """
        return True


class ConditionalMultiInputConnector(MultiInputConnector):
    """A variant of the :class:`~connectors.MultiInputConnector`, in which it
    depends on externally defined conditions, if the value changes are announced
    to the observing output connectors, and if the output connectors are notified
    about value changes.

    The conditions are defined as methods, which are called before sending an
    announcement or a notification to the observing output connectors and which
    shall return a Boolean, that is True, if the output connector shall be informed,
    about the (expected) value change, or False, if the output shall be sent a
    cancellation notice.
    """

    def __init__(self, instance, method,                            # pylint: disable=too-many-arguments; this constructor may be complicated, since the class is instantiated through the decorators, which have a much simpler API
                 remove_method, replace_method,
                 observers, announce_condition, notify_condition,
                 laziness, parallelization, executor):
        """
        :param instance: the instance of which the method is replaced by this connector
        :param method: the unbound method, that is replaced by this connector
        :param remove_method: an unbound method, that is used to remove data, that
                              has been added through this connector
        :param replace_method: an unbound method, that is used to replace data,
                               that has been added through this connector
        :param observers: the names of output methods that are affected by passing a value to this connector
        :param announce_condition: a method, that defines the condition for the
                                   announcements to the observing output connectors.
                                   This method must accept the data ID of the changed
                                   output connector as an argument in addition to
                                   ``self``
        :param notify_condition: a method, that defines the condition for the
                                 notifications to the observing output connectors.
                                 This method must accept the data ID and the new
                                 input value as an argument in addition to ``self``
        :param laziness: a flag from the :class:`connectors.Laziness` enum. See
                         the :meth:`set_laziness` method for details
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :meth:`set_parallelization` method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :meth:`~connectors.connectors.MultiInputConnector.set_executor`
                         method for details
        """
        MultiInputConnector.__init__(self, instance, method,
                                     remove_method, replace_method,
                                     observers,
                                     laziness, parallelization, executor)
        self.__announce_condition = announce_condition
        self.__notify_condition = notify_condition
        self.__pending_notification_condition_checks = []

    def _announce(self, connector, non_lazy_inputs):
        """This method is to notify this input connector, when a connected output
        connector can produce updated data.

        :param connector: the output connector whose value is about to change
        :param non_lazy_inputs: a :class:`~connectors._common._non_lazy_inputs.NonLazyInputs`
                                instance to which input connectors can be appended,
                                if they request an immediate re-computation (see
                                the :meth:`~connectors.connectors.MultiInputConnector.set_laziness`
                                method for more about lazy execution)
        """
        if self.__announce_condition(self._instance(), self._connections[connector]):
            super()._announce(connector, non_lazy_inputs)

    def _conditional_observer_notification(self, data_id, *args, **kwargs):
        """Notifies the observing output connectors, that this input has changed
        the instance's state.

        This method can be overridden by derived classes, for example to implement
        a conditional notification of the outputs.

        :param `*args,**kwargs`: the parameters, with which the replaced setter
                                method has been called (excluding ``self``)
        """
        value = common.get_first_argument(self._method, *args, **kwargs)
        if self.__notify_condition(self._instance(), data_id, value):
            super()._conditional_observer_notification(data_id, value)
        else:
            super()._cancel(None)

    def _add_to_notification_condition_checks(self, data_id, value):
        """A protected helper method to determine, if the observing output connectors
        have to be notified at the end of the :meth:`~connectors.MultiInputConnector._request`
        method.
        """
        self.__pending_notification_condition_checks.append((data_id, value))

    def _check_notification_condition(self):
        """A protected helper method to determine, if the observing output connectors
        have to be notified at the end of the :meth:`~connectors.MultiInputConnector._request`
        method.
        """
        result = False
        for data_id, value in self.__pending_notification_condition_checks:
            result = result or self.__notify_condition(self._instance(), data_id, value)
        self.__pending_notification_condition_checks.clear()
        return result



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains the :class:`Multiplexer` class."""

import connectors

__all__ = ("Multiplexer",)


class Multiplexer:
    """A class, that routes one of arbitrarily many inputs to its output.

    Usage Example: (the assignments ``_ = ...`` are only to make :mod:`doctest`
    ignore the return values of the operations.)

    >>> import connectors
    >>> # Create some test objects
    >>> test1 = connectors.blocks.Passthrough(data="One")
    >>> test2 = connectors.blocks.Passthrough(data="Two")
    >>> multiplexer = connectors.blocks.Multiplexer()
    >>> # Connect the test objects to the multiplexer
    >>> # Note, how the input of the input connector of the multiplexer is accessed like a dictionary
    >>> _ = multiplexer.input["1"].connect(test1.output)
    >>> _ = test2.output.connect(multiplexer.input[2])    # the order, which is connected to which is not important
    >>> # Select the output with the same selector, that has been passed as key, during connecting
    >>> _ = multiplexer.select("1")
    >>> multiplexer.output()
    'One'
    >>> _ = multiplexer.select(2)
    >>> multiplexer.output()
    'Two'
    """

    def __init__(self, selector=None):
        """
        :param selector: the selector of the input, that shall be routed to the output
        """
        self.__selector = selector
        self.__data = connectors.MultiInputData()

    @connectors.Output()
    def output(self):
        """Returns the value from the selected input.
        If no data is found for the given selector, the first input data set, that
        was added, is returned.

        :returns: the selected input object
        """
        if self.__selector in self.__data:
            return self.__data[self.__selector]
        else:
            return None

    @connectors.Input("output")
    def select(self, selector):
        """Specifies, which input shall be routed to the output.

        :param selector: the selector of the input, that shall be routed to the output
        :returns: the Multiplexer instance
        """
        self.__selector = selector
        return self

    @connectors.MultiInput("output")
    def input(self, data):
        """Specifies an input object, that can be selected to be routed to the output.
        When connecting to this method, the selector, by which the given connection
        can be selected, can be specified with the :meth:`~connectors.connectors.MultiInputConnector.__getitem__`
        overload::

           multiplexer.input[selector].connect(generator.output())

        :param data: the input object
        :returns: an ID, that can be used as a selector value t
        """
        return self.__data.add(data)

    @input.remove
    def remove(self, data_id):
        """Is used to remove an input object from the collection, when the respective
        connector is disconnected from the input.

        :param data_id: the data_id under which the input object is stored
        :returns: the Multiplexer instance
        """
        del self.__data[data_id]
        return self

    @input.replace
    def replace(self, data_id, data):
        """Is used to replace an input object, when the respective connected
        connector has produced a new one.

        :param data_id: the data_id under which the input object is stored
        :param data: the new input object
        :returns: ``data_id``, because the ID does not change for the replaced data
        """
        self.__data[data_id] = data
        return data_id

    @input.notify_condition
    def __input_condition(self, data_id, value):   # pylint: disable=unused-argument; notify condition methods must accept the new value
        """Prevents, that the output is notified about changes of an input, that
        is not currently selected.
        """
        return data_id == self.__selector



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains the :class:`Passthrough` class."""

import connectors

__all__ = ("Passthrough",)


class Passthrough:
    """A trivial processing block, that simply passes its input value to its output.
    Instances of this can be useful to distribute a single parameter to multiple
    inputs, if this parameter is used in several places in a processing chain.
    """

    def __init__(self, data=None):
        """
        :param data: the input object
        """
        self.__data = data

    @connectors.Output(caching=False)
    def output(self):
        """Returns the object, that has been passed with the :meth:`~Passthrough.input` method.

        :returns: the given object
        """
        return self.__data

    @connectors.Input("output")
    def input(self, data):
        """Specifies the input object.

        :param data: the input object
        :returns: the :class:`Passthrough` instance
        """
        self.__data = data
        return self



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains the OutputProxy class"""

from ..connectors import OutputConnector
from ._baseclasses import ConnectorProxy

__all__ = ("OutputProxy",)


class OutputProxy(ConnectorProxy):
    """A proxy class for output connectors.
    Connector proxies are returned by the connector decorators, while the methods
    are replaced by the actual connectors. Think of a connector proxy like of a
    bound method, which is also created freshly, whenever a method is accessed.
    The actual connector only has a weak reference to its instance, while this
    proxy has a hard reference, but mimics the connector in almost every other
    way. This makes constructions like ``value = Class().connector()`` possible.
    Without the proxy, the instance of the class would be deleted before the connector
    method is called, so that the weak reference of the connector would be expired
    during its call.
    """

    def __init__(self, instance, method, caching, parallelization, executor):
        """
        :param instance: the instance in which the method is replaced by this connector proxy
        :param method: the unbound method that is replaced by this connector proxy
        :param caching: True, if caching shall be enabled, False otherwise. See
                        the :meth:`set_caching` method for details
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :meth:`set_parallelization` method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :class:`~connectors.connectors.OutputConnector`'s
                         :meth:`~connectors.connectors.OutputConnector.set_executor`
                         method for details
        """
        ConnectorProxy.__init__(self, instance, method, parallelization, executor)
        self.__caching = caching

    def set_caching(self, caching):
        """Specifies, if the result value of this output connector shall be cached.
        If caching is enabled and the result value is retrieved (e.g. through a
        connection or by calling the connector), the cached value is returned and
        the replaced getter method is not called unless the result value has to
        be re-computed, because an observed setter method has changed a parameter
        for the computation. In this case, the getter method is only called once,
        independent of the number of connections through which the result value
        has to be passed.

        :param caching: True, if caching shall be enabled, False otherwise
        """
        self._get_connector().set_caching(caching)

    def _create_connector(self, instance, method, parallelization, executor):
        """Creates and returns the output connector.

        :param instance: the instance in which the method is replaced by the connector
        :param method: the unbound method that is replaced by the connector
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :meth:`set_parallelization` method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :class:`~connectors.connectors.OutputConnector`'s
                         :meth:`~connectors.connectors.OutputConnector.set_executor`
                         method for details
        :returns: an :class:`~connectors.connectors.OutputConnector` instance
        """
        return OutputConnector(instance=instance,
                               method=method,
                               caching=self.__caching,
                               parallelization=parallelization,
                               executor=executor)

    def _announce(self, connector, non_lazy_inputs):
        """This method is to notify this output connector, when an observed input
        connector (a setter from self._instance) can retrieve updated data.

        :param connector: the input connector, which is about to change a value
        :param non_lazy_inputs: a :class:`~connectors._common._non_lazy_inputs.NonLazyInputs`
                                instance to which input connectors can be appended,
                                if they request an immediate re-computation (see
                                the :class:`~connectors._connectors._base_classes.InputConnector`'s
                                :meth:`~connectors._connectors._base_classes.InputConnector.set_laziness`
                                method for more about lazy execution)
        """
        # nothing to do for a proxy

    def _notify(self, connector):
        """This method is to notify this output connector, when an observed input
        connector (a setter from self._instance) has retrieved updated data.

        :param connector: the input connector, which has changed a value
        :param non_lazy_inputs: a :class:`~connectors._common._non_lazy_inputs.NonLazyInputs`
                                instance to which input connectors can be appended,
                                if they request an immediate re-computation (see
                                the :class:`~connectors._connectors._base_classes.InputConnector`'s
                                :meth:`~connectors._connectors._base_classes.InputConnector.set_laziness`
                                method for more about lazy execution)
        """
        # nothing to do for a proxy



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""The connectors package implements decorators for setter and getter methods, so
they can be connected to form a processing chain, in which the result values are
automatically updated, whenever a parameter has been changed. This is similar to
a pipes and filters architecture.
The computations can also be parallelized automatically with threads and processes.
"""

# pylint: disable=wildcard-import;    wildcard imports shall be allowed here, since all submodules have their __all__-variables defined

from ._common import Laziness, Parallelization, executor
from ._decorators import *
from ._helper import *
from ._macro import *

from . import blocks



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Defines a container class for tracking non-lazy input connectors"""

import asyncio

__all__ = ("NonLazyInputs",)


class NonLazyInputs(set):
    """A subclass of :class:`set`, that is used internally to track the non-lazy
    input connectors, that request an immediate re-computation of the processing
    chain.
    """

    def __init__(self, situation):
        """
        :param situation: a flag from the :class:`~connectors.Laziness` enumeration
                          to which the laziness of the connectors is compared in
                          order to decide, if it shall be added to this :class:`set`.
        """
        set.__init__(self)
        self.__situation = situation

    def add(self, connector, laziness):
        """Adds a connector to this container, if its laziness is low enough to
        cause immediate execution.

        :param connector: the :class:`~connectors._connectors._baseclasses.InputConnector`
                          instance, that shall be added
        :param laziness: the laziness setting of that connector as a flag from the
                         :class:`~connectors.Laziness` enumeration
        """
        if laziness >= self.__situation:
            set.add(self, connector)

    def execute(self, executor):
        """Executes the necessary computations, that are requested by the non-lazy
        input connectors.

        :param executor: the :class:`connectors._common._executors.Executor` instance,
                         that manages the executions
        """
        if self:
            tasks = asyncio.wait([i._request(executor) for i in self])
            executor.run_until_complete(tasks)



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains the SingleInputConnector class"""

from .. import _common as common
from ._baseclasses import InputConnector

__all__ = ("SingleInputConnector", "ConditionalSingleInputConnector")


class SingleInputConnector(InputConnector):
    """A Connector-class that replaces setter methods, so they can be used to connect
    different objects in a processing chain."""

    def __init__(self, instance, method, observers, laziness, parallelization, executor):
        """
        :param instance: the instance of which the method is replaced by this connector
        :param method: the unbound method, that is replaced by this connector
        :param observers: the names of output methods that are affected by passing a value to this connector
        :param laziness: a flag from the :class:`connectors.Laziness` enum. See
                         the :meth:`~connectors.connectors.SingleInputConnector.set_laziness`
                         method for details
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :meth:`~connectors.connectors.SingleInputConnector.set_parallelization`
                                method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :meth:`~connectors.connectors.SingleInputConnector.set_executor`
                         method for details
        """
        InputConnector.__init__(self, instance, method, laziness, parallelization, executor)
        self.__observers = common.resolve_observers(instance=instance, observers=observers)
        self.__announcement = None
        self.__notification = None
        self.__notification_is_valid = False
        self.__running = False              # is used to prevent, that the setter is executed multiple times for the same changes
        self.__computable = common.Event()  # is set, when there is no pending announcement
        self.__computable.set()

    def __call__(self, *args, **kwargs):
        """By making the object callable, it mimics the replaced method.
        This method also notifies the output method that are affected by this call (observers).

        :param `*args,**kwargs`: parameters with which the replaced method shall be called
        :returns: the return value of the replaced method
        """
        # announce the value change
        non_lazy_inputs = common.NonLazyInputs(common.Laziness.ON_ANNOUNCE)
        for o in self.__observers:
            o._announce(self, non_lazy_inputs)
        # call the replaced method
        result = self._method(self._instance(), *args, **kwargs)
        self.__announcement = None
        self.__notification = None
        self.__notification_is_valid = False
        # notify observers about the value change
        self._conditional_observer_notification(*args, **kwargs)
        # execute the non-lazy inputs
        non_lazy_inputs.execute(self._executor)
        # return the result of the method call
        return result

    def _connect(self, connector):
        """This method is called from an :class:`~connectors.connectors.OutputConnector`,
        when it is  being connected to this :class:`~connectors.connectors.SingleInputConnector`.

        :param connector: the :class:`~connectors.connectors.OutputConnector` instance
                          to which this connector shall be connected
        :returns: yields ``self`` (see the :class:`~connectors.connectors.MacroInputConnector`,
                  where :meth:`~connectors.connectors.MacroInputConnector._connect`
                  yields all the :class:`~connectors.connectors.SingleInputConnector`
                  instances that it exports)
        """
        yield self
        non_lazy_inputs = common.NonLazyInputs(situation=common.Laziness.ON_CONNECT)
        self._announce(connector, non_lazy_inputs=non_lazy_inputs)
        non_lazy_inputs.execute(self._executor)

    def _disconnect(self, connector):   # pylint: disable=unused-argument; this method has to be compatible with other input connectors
        """This method is called from an :class:`~connectors.OutputConnector`,
        when it is  being disconnected from this :class:`~connectors.SingleInputConnector`.

        :param connector: the output connector from which this connector shall be disconnected
        :returns: yields self (see the :class:`~connectors.MacroInputConnector`,
                  where :meth:`~connectors.MacroInputConnector._disconnect` yields
                  all the :class:`~connectors.SingleInputConnector`s that it exports)
        """
        self._executor.run_coroutine(self._request(self._executor))
        yield self

    def _announce(self, connector, non_lazy_inputs):
        """This method is to inform this input connector, when a connected output
        connector can produce updated data.

        :param connector: the output connector whose value is about to change
        :param non_lazy_inputs: a :class:`~connectors._common._non_lazy_inputs.NonLazyInputs`
                                instance to which input connectors can be appended,
                                if they request an immediate re-computation (see
                                the :meth:`~connectors.connectors.SingleInputConnector.set_laziness`
                                method for more about lazy execution)
        """
        self.__announcement = connector
        self.__notification = None
        self.__notification_is_valid = False
        self.__computable.clear()
        common.forward_announcement(connector=self,
                                    observers=self.__observers,
                                    non_lazy_inputs=non_lazy_inputs,
                                    laziness=self._laziness)

    async def _notify(self, connector, value, executor):    # pylint: disable=unused-argument; this method has to be compatible with other input connectors
        """This method is to notify this input connector, when a connected output
        connector has produced updated data.

        :param connector: the output connector whose value has changed
        :param value: the updated data from the output connector
        :param executor: the :class:`~connectors._common._executors.Executor`
                         instance, which managed the computation of the output
                         connector, and which shall be used for the computation
                         of this connector, in case it is not lazy.
        """
        self.__notification = value
        self.__notification_is_valid = True
        self.__announcement = None
        self.__computable.set()
        if self._laziness == common.Laziness.ON_NOTIFY:
            await self._request(executor)

    def _cancel(self, connector):   # pylint: disable=unused-argument; this method has to be compatible with other input connectors
        """Notifies this input connector, that an announced value change is not
        going to happen.

        :param connector: the output connector whose value change is canceled
        """
        self.__announcement = None
        self.__computable.set()
        for o in self.__observers:
            o._cancel(self)     # pylint: disable=protected-access; the _cancel method is meant to be called from other connectors, but not from outside this package

    async def _request(self, executor):
        """This method retrieves the updated data from the connected output connector
        and recomputes this connector (if necessary).
        It is called by the output connectors, that are affected by this connector,
        (observers) when their values have to be computed.

        :param executor: the :class:`~connectors._common._executors.Executor`
                         instance, that manages the current computations
        """
        if not self.__running:
            self.__running = True
            try:
                # wait for the announced value change
                if self.__announcement is not None:
                    await self.__announcement._request(executor)
                    await self.__computable.wait(executor)
                # execute the setter
                if self.__notification_is_valid:
                    notification = self.__notification
                    await executor.run_method(self._parallelization, self._method, self._instance(), notification)
                    self.__notification_is_valid = False
                    self.__notification = None
                    # notify the observers
                    self._conditional_observer_notification(notification)
            finally:
                self.__running = False

    def _conditional_observer_notification(self, *args, **kwargs):  # pylint: disable=unused-argument; this method has to be compatible with other input connectors
        """Notifies the observing output connectors, that this input has changed
        the instance's state.

        This method can be overridden by derived classes, for example to implement
        a conditional notification of the outputs.

        :param `*args,**kwargs`: the parameters, with which the replaced setter
                                method has been called (excluding ``self``)
        """
        for o in self.__observers:
            o._notify(self)


class ConditionalSingleInputConnector(SingleInputConnector):
    """A variant of the :class:`~connectors.SingleInputConnector`, in which it
    depends on externally defined conditions, if the value changes are announced
    to the observing output connectors, and if the output connectors are notified
    about value changes.

    The conditions are defined as methods, which are called before sending an
    announcement or a notification to the observing output connectors and which
    shall return a Boolean, that is True, if the output connector shall be informed,
    about the (expected) value change, or False, if the output shall be sent a
    cancellation notice.
    """

    def __init__(self, instance, method, observers,
                 announce_condition, notify_condition,
                 laziness, parallelization, executor):
        """
        :param instance: the instance of which the method is replaced by this connector
        :param method: the unbound method, that is replaced by this connector
        :param observers: the names of output methods that are affected by passing a value to this connector
        :param announce_condition: a method, that defines the condition for the
                                   announcements to the observing output connectors.
                                   This method must not require any arguments apart
                                   from ``self``
        :param notify_condition: a method, that defines the condition for the
                                 notifications to the observing output connectors.
                                 This method must accept the new input value as
                                 an argument in addition to ``self``
        :param laziness: a flag from the :class:`connectors.Laziness` enum. See
                         the :meth:`set_laziness` method for details
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :meth:`set_parallelization` method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :meth:`~connectors.connectors.SingleInputConnector.set_executor`
                         method for details
        """
        SingleInputConnector.__init__(self, instance, method, observers, laziness, parallelization, executor)
        self.__announce_condition = announce_condition
        self.__notify_condition = notify_condition

    def _announce(self, connector, non_lazy_inputs):
        """This method is to inform this input connector, when a connected output
        connector can produce updated data.

        :param connector: the output connector whose value is about to change
        :param non_lazy_inputs: a :class:`~connectors._common._non_lazy_inputs.NonLazyInputs`
                                instance to which input connectors can be appended,
                                if they request an immediate re-computation (see
                                the :meth:`~connectors.connectors.SingleInputConnector.set_laziness`
                                method for more about lazy execution)
        """
        if self.__announce_condition(self._instance()):
            super()._announce(connector, non_lazy_inputs)

    def _conditional_observer_notification(self, *args, **kwargs):
        """Notifies the observing output connectors, that this input has changed
        the instance's state, if the condition is met.

        :param `*args,**kwargs`: the parameters, with which the replaced setter
                                method has been called (excluding **self**)
        """
        value = common.get_first_argument(self._method, *args, **kwargs)
        if self.__notify_condition(self._instance(), value):
            super()._conditional_observer_notification(value)
        else:
            super()._cancel(None)



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains the :class:`MultiinputItem` class"""

__all__ = ("MultiinputItem",)

from ._flags import Laziness
from ._non_lazy_inputs import NonLazyInputs


class MultiinputItem:
    """An object, that is returned by the :meth:`~connectors.MultiInputConnector.__getitem__` overload.

    It simulates the behavior of a single-input connector, so it is possible to
    use a multi-input connector as arbitrarily many single-inputs.
    """

    def __init__(self, connector, instance, replace_method, key, observers, executor):
        """
        :param connector: the multi-input connector
        :param instance: the instance of which the method was replaced by the
                         multi-input connector
        :param replace_method: an unbound method, that is used to replace data,
                               that has been added through the multi-input connector
        :param key: the key with which the multi-input has been accessed.
        :param observers: a sequence of output connectors, that observe the
                          multi-input connector's value changes.
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that is used, when calling the instance of this class.
        """
        self.__connector = connector
        self.__instance = instance
        self.__replace = replace_method
        self.__key = key
        self.__observers = observers
        self.__executor = executor

    def __call__(self, *args, **kwargs):
        """Calls the given replace-method

        :param `*args,**kwargs`: arguments for the replace-method
        :returns: the instance of which the method was replaced by the multi-input connector
        """
        if self.__observers:
            non_lazy_inputs = NonLazyInputs(Laziness.ON_ANNOUNCE)
            for o in self.__observers:
                o._announce(self.__connector, non_lazy_inputs)
            self.__replace(self.__instance, self.__key, *args, **kwargs)
            self.__connector._conditional_observer_notification(self.__key, *args, **kwargs)    # pylint: disable=protected-access; this is not super clean, but since it all stays within the context of a multi-input connector...
            non_lazy_inputs.execute(self.__executor)
        else:
            self.__replace(self.__instance, self.__key, *args, **kwargs)
        return self.__instance

    def connect(self, connector):
        """Connects this virtual single-input to an output.

        :param connector: the connector, to which this connector shall be connected
        :returns: the instance of which the method was replaced by the multi-input connector
        """
        connector.connect(self)
        return self.__instance

    def disconnect(self, connector):
        """Disconnects this virtual single-input from an output, to which is has been connected..

        :param connector: the connector, from which this connector shall be disconnected
        :returns: the instance of which the method was replaced by the multi-input connector
        """
        connector.disconnect(self)
        return self.__instance

    def _connect(self, connector):
        """This method is called from an :class:`OutputConnector`, when it is being
        connected to this virtual single-input.

        :param connector: the connector to which this connector shall be connected
        :returns: yields the instance, of which a method has been replaced by the multi-input connector
        """
        return self.__connector._get_connector()._connect(connector, key=self.__key)

    def _disconnect(self, connector):
        """This method is called from an :class:`OutputConnector`, when it is being
        disconnected from this virtual single-input.

        :param connector: the connector from which this connector shall be disconnected
        :returns: yields the instance, of which a method has been replaced by the multi-input connector
        """
        return self.__connector._disconnect(connector)



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains supplementary classes for the implementation of the remove and replace
methods of multi-input connectors.
"""

import functools
from ._flags import Laziness
from ._non_lazy_inputs import NonLazyInputs

__all__ = ("MultiInputAssociateDescriptor", "MultiInputAssociateProxy",)


class MultiInputAssociateDescriptor:
    """A descriptor class for associating remove and replace methods with their
    multi-input.
    Instances of this class are created by the decorator methods :meth:`~connectors.MultiInput.remove`
    and :meth:`~connectors.MultiInput.replace`.
    """

    def __init__(self, method, observers, parallelization, executor):
        """
        :param method: the unbound method, that is wrapped
        :param observers: the names of output methods that are affected by passing
                          a value to the multi-input connector.
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :class:`~connectors.connectors.MultiInputConnector`'s
                                :meth:`~connectors.connectors.MultiInputConnector.set_parallelization`
                                method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :class:`~connectors.connectors.MultiInputConnector`'s
                         :meth:`~connectors.connectors.MultiInputConnector.set_executor`
                         method for details
        """
        self.__method = method
        self.__observers = observers
        self.__parallelization = parallelization
        self.__executor = executor

    def __get__(self, instance, instance_type):
        """Is called, when the decorated method is accessed.

        :param instance: the instance of which a method shall be replaced
        :param instance_type: the type of the instance
        :returns: a :class:`connectors._common._multiinput_associate.MultiInputAssociateProxy`
                  instance, that mimics the decorated method and adds the connector
                  functionality
        """
        return MultiInputAssociateProxy(instance=instance,
                                        method=self.__method,
                                        observers=self.__observers,
                                        executor=self.__executor)


class MultiInputAssociateProxy:
    """A proxy class for remove or replace methods of multi-input connectors.
    Connector proxies are returned by the connector decorators, while the methods
    are replaced by the actual connectors. Think of a connector proxy like of a
    bound method, which is also created freshly, whenever a method is accessed.
    The actual connector only has a weak reference to its instance, while this
    proxy has a hard reference, but mimics the connector in almost every other
    way. This makes constructions like ``value = Class().connector()`` possible.
    Without the proxy, the instance of the class would be deleted before the connector
    method is called, so that the weak reference of the connector would be expired
    during its call.
    """

    def __init__(self, instance, method, observers, executor):
        """
        :param instance: the instance in which the method is replaced by the multi-input connector proxy
        :param method: the unbound method, that is replaced by this proxy (the remove or replace method)
        :param observers: the names of output methods that are affected by passing
                          a value to the multi-input connector proxy
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :meth:`~connectors.connectors.MultiInputConnector.set_executor`
                         method for details
        """
        self.__instance = instance
        self.__method = method
        self.__observers = observers
        self.__executor = executor
        functools.update_wrapper(self, method)

    def __call__(self, *args, **kwargs):
        """Executes the replaced method and notifies the observing output connectors.
        :param `*args,**kwargs`: possible arguments for the replaced method
        :returns: the return value of the method, that has been replaced by this
        """
        instance = self.__instance
        # announce the value change
        non_lazy_inputs = NonLazyInputs(Laziness.ON_ANNOUNCE)
        for o in self.__observers:
            getattr(instance, o)._announce(self, non_lazy_inputs)
        # call the replaced method
        result = self.__method(self.__instance, *args, **kwargs)
        # notify observers about the value change
        for o in self.__observers:
            getattr(instance, o)._notify(self)
        # execute the non-lazy inputs
        non_lazy_inputs.execute(self.__executor)
        # return the result of the method call
        return result



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains the SingleInputProxy class"""

from ..connectors import SingleInputConnector, ConditionalSingleInputConnector
from .. import _common as common
from ._baseclasses import ConnectorProxy

__all__ = ("SingleInputProxy",)


class SingleInputProxy(ConnectorProxy):
    """A proxy class for input connectors.
    Connector proxies are returned by the connector decorators, while the methods
    are replaced by the actual connectors. Think of a connector proxy like of a
    bound method, which is also created freshly, whenever a method is accessed.
    The actual connector only has a weak reference to its instance, while this
    proxy has a hard reference, but mimics the connector in almost every other
    way. This makes constructions like ``value = Class().connector()`` possible.
    Without the proxy, the instance of the class would be deleted before the connector
    method is called, so that the weak reference of the connector would be expired
    during its call.
    """

    def __init__(self, instance, method,
                 observers, announce_condition, notify_condition,
                 laziness, parallelization, executor):
        """
        :param instance: the instance in which the method is replaced by this connector proxy
        :param method: the unbound method that is replaced by this connector proxy
        :param observers: the names of output methods that are affected by passing a value to this connector proxy
        :param announce_condition: a method, that defines the condition for the
                                   announcements to the observing output connectors.
                                   This method must not require any arguments apart
                                   from ``self``
        :param notify_condition: a method, that defines the condition for the
                                 notifications to the observing output connectors.
                                 This method must accept the new input value as
                                 an argument in addition to ``self``
        :param laziness: a flag from the :class:`connectors.Laziness` enum. See
                         the :meth:`set_laziness` method for details
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :meth:`set_parallelization` method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :class:`~connectors.connectors.SingleInputConnector`'s
                         :meth:`~connectors.connectors.SingleInputConnector.set_executor`
                         method for details
        """
        ConnectorProxy.__init__(self, instance, method, parallelization, executor)
        self._observers = observers
        self._laziness = laziness
        self._announce_condition = announce_condition
        self._notify_condition = notify_condition

    def __call__(self, *args, **kwargs):
        """Executes the replaced method and notifies the observing output connectors.
        :param `*args,**kwargs`: possible arguments for the replaced method
        """
        instance = self._get_instance()
        # announce the value change
        non_lazy_inputs = common.NonLazyInputs(common.Laziness.ON_ANNOUNCE)
        for o in self._observers:
            getattr(instance, o)._announce(self, non_lazy_inputs)
        # call the replaced method
        result = ConnectorProxy.__call__(self, *args, **kwargs)
        # notify observers about the value change
        for o in self._observers:
            getattr(instance, o)._notify(self)
        # execute the non-lazy inputs
        non_lazy_inputs.execute(self._executor)
        # return the result of the method call
        return result

    def set_laziness(self, laziness):
        """Configures the lazy execution of the connector.
        Normally the connectors are executed lazily, which means, that any computation
        is only started, when the result of a processing chain is requested. For
        certain use cases it is necessary to disable this lazy execution, though,
        so that the values are updated immediately as soon as new data is available.
        There are different behaviors for the (non) lazy execution, which are
        described in the :class:`connectors.Laziness` enum.

        :param laziness: a flag from the :class:`connectors.Laziness` enum
        """
        self._get_connector().set_laziness(laziness)

    def _connect(self, connector):
        """A method for connecting another connector to this connector.
        Calling this method causes the actual connector, that is represented by
        this proxy, to be created and the method to be replaced. This proxy is no
        longer used after this.
        This method is meant to be used internally in the *Connectors* package.

        :param connector: the connector to which this connector shall be connected
        """
        return self._get_connector()._connect(connector)

    def _create_connector(self, instance, method, parallelization, executor):
        """Creates and returns the input connector.

        :param instance: the instance in which the method is replaced by the connector
        :param method: the unbound method that is replaced by the connector
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :meth:`set_parallelization` method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :class:`~connectors.connectors.SingleInputConnector`'s
                         :meth:`~connectors.connectors.SingleInputConnector.set_executor`
                         method for details
        :returns: an :class:`SingleInputConnector` instance
        """
        if self._announce_condition is None and self._notify_condition is None:
            return SingleInputConnector(instance=instance,
                                        method=method,
                                        observers=self._observers,
                                        laziness=self._laziness,
                                        parallelization=parallelization,
                                        executor=executor)
        else:
            announce_condition, notify_condition = common.select_condition_methods(self._announce_condition,
                                                                                   self._notify_condition)
            return ConditionalSingleInputConnector(instance=instance,
                                                   method=method,
                                                   observers=self._observers,
                                                   announce_condition=announce_condition,
                                                   notify_condition=notify_condition,
                                                   laziness=self._laziness,
                                                   parallelization=parallelization,
                                                   executor=executor)



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains the decorator and connector classes for the macro output connector."""

import functools
from ._baseclass import MacroDecorator

__all__ = ("MacroOutput",)


class MacroOutput(MacroDecorator):
    """A decorator to replace a method with a macro output connector.

    Macro connectors are useful, when a processing network shall be encapsulated
    in a class. In such a case, macro output connectors are used to export output
    connectors from the internal processing network to be accessible as connectors
    of the encapsulating class's instance.

    The decorated method must not take any parameters and return the output connector
    from the internal processing network, that it exports.

    The resulting connector will behave like an output connector, which is very
    different from the decorated method:

    - the connector takes no argument.
    - when called, the connector returns the result of the exported connector.
    """

    def __get__(self, instance, instance_type):
        """Is called, when the decorated method is accessed.

        :param instance: the instance of which a method shall be replaced
        :param instance_type: the type of the instance
        :returns: a :class:`MacroInputConnector` instance
        """
        return MacroOutputConnector(instance=instance, method=self._method)


class MacroOutputConnector:
    """A Connector-class that exports an output connector from an internal processing
    network to the API of the class, that encapsulates the network.
    """

    def __init__(self, instance, method):
        """
        :param instance: the instance in which the method is replaced by this connector
        :param method: the unbound method, that is replaced by this connector
        """
        self.__instance = instance
        self.__method = method
        functools.update_wrapper(self, method)

    def __call__(self):
        """Calls the output connector, that is exported by this.

        :returns: the return value from the call
        """
        return self.__method(self.__instance)()

    def set_caching(self, caching):
        """Specifies, if the result value of this output connector shall be cached.
        If caching is enabled and the result value is retrieved (e.g. through a
        connection or by calling the connector), the cached value is returned and
        the replaced getter method is not called unless the result value has to
        be re-computed, because an observed setter method has changed a parameter
        for the computation. In this case, the getter method is only called once,
        independent of the number of connections through which the result value
        has to be passed.

        :param caching: True, if caching shall be enabled, False otherwise
        """
        self.__method(self.__instance).set_caching(caching)

    def set_parallelization(self, parallelization):
        """Specifies, if and how the execution of this connector can be parallelized.
        The choices are no parallelization, the execution in a separate thread
        and the execution in a separate process.
        This method specifies a hint, which level of parallelization is possible
        with the connector. If the executor of the connector, through which the
        computation is started, does not support the specified level, the next simpler
        one will be chosen. E.g. if a connector can be parallelized in a separate
        process, but the executor only allows threads or sequential execution, the
        connector will be executed in a separate thread.

        :param parallelization: a flag from the :class:`connectors.Parallelization` enum
        """
        self.__method(self.__instance).set_parallelization(parallelization)

    def set_executor(self, executor):
        """Sets the executor, which handles the computations, when the data is
        retrieved through this connector.
        An executor can be created with the :func:`connectors.executor` function. It
        manages the order and the parallelization of the computations, when updating
        the data in a processing chain.
        If multiple connectors in a processing chain need to be computed, the
        executor of the connector, which started the computations, is used for
        all computations.

        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function
        """
        self.__method(self.__instance).set_executor(executor)

    def connect(self, connector):
        """Connects the exported output connector to the given input.

        :param connector: the input connector to which the exported connector shall be connected
        :returns: the instance of which this connector has replaced a method
        """
        self.__method(self.__instance).connect(connector)
        return self.__instance

    def disconnect(self, connector):
        """Disconnects the exported output connector from the given input.

        :param connector: the input connector from which the exported connector shall be disconnected
        :returns: the instance of which this connector has replaced a method
        """
        self.__method(self.__instance).disconnect(connector)
        return self.__instance



from setuptools import setup, find_packages

with open("README.rst") as f:
    long_description = f.read()

setup(name="connectors",
      version="3.0",
      description="A package for connecting objects to form a processing chain",
      long_description=long_description,
      url="https://github.com/JonasSC/Connectors",
      author="Jonas Schulte-Coerne",
      license="GNU Lesser General Public License v3 or later (LGPLv3+)",
      classifiers=["Development Status :: 5 - Production/Stable",
                   "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
                   "Programming Language :: Python :: 3.6",
                   "Intended Audience :: Developers",
                   "Intended Audience :: Science/Research",
                   "Topic :: Software Development :: Libraries :: Application Frameworks"],
      keywords="parallel observer pipes-and-filters",
      packages=find_packages(exclude=["documentation", "tests", "docs"]),
      install_requires=[],
      extras_require={"dev": ["pytest-cov", "pylint", "flake8"],
                      "test": ["pytest", "numpy"],
                      "doc": ["sphinx_rtd_theme"]},
      project_urls={"Documentation": "https://connectors.readthedocs.io/en/latest/",
                    "Bug Reports": "https://github.com/JonasSC/Connectors/issues",
                    "Source": "https://github.com/JonasSC/Connectors"})



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains the :class:`Output` class for decorating getter methods"""

from connectors._common import Parallelization
from .._proxies import OutputProxy
from ._baseclasses import ConnectorDecorator, default_executor

__all__ = ("Output",)


class Output(ConnectorDecorator):
    """A decorator, that marks a method as an output connector.
    These connections can be used to automatically update a processing chain
    when a value has changed.
    The decorated method must not take any arguments.
    """

    def __init__(self,
                 caching=True,
                 parallelization=Parallelization.get_default_output_parallelization(),
                 executor=default_executor):
        """
        :param caching: True, if caching shall be enabled, False otherwise. See
                        the :class:`~connectors.connectors.OutputConnector`'s
                        :meth:`~connectors.connectors.OutputConnector.set_caching`
                        method for details
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :class:`~connectors.connectors.OutputConnector`'s
                                :meth:`~connectors.connectors.OutputConnector.set_parallelization`
                                method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :class:`~connectors.connectors.OutputConnector`'s
                         :meth:`~connectors.connectors.OutputConnector.set_executor`
                         method for details
        """
        ConnectorDecorator.__init__(self, parallelization, executor)
        self.__caching = caching

    def __get__(self, instance, instance_type):
        """Is called, when the decorated method is accessed.

        :param instance: the instance of which a method shall be replaced
        :param instance_type: the type of the instance
        :returns: an :class:`~connectors.proxies.OutputProxy` instance, that mimics
                  the decorated method and adds the connector functionality
        """
        return OutputProxy(instance=instance,
                           method=self._method,
                           caching=self.__caching,
                           parallelization=self._parallelization,
                           executor=self._executor)



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains the :class:`~connectors.connectors.OutputConnector` class"""

import asyncio
import weakref
from .. import _common as common
from ._baseclasses import Connector

__all__ = ("OutputConnector",)


class OutputConnector(Connector):
    """A Connector-class that replaces getter methods, so they can be used to
    connect different objects.
    """

    def __init__(self, instance, method, caching, parallelization, executor):
        """
        :param instance: the instance of which the method is replaced by this connector
        :param method: the unbound method that is replaced by this connector
        :param caching: True, if caching shall be enabled, False otherwise. See
                        the :meth:`~connectors.connectors.OutputConnector.set_caching`
                        method for details
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :meth:`~connectors.connectors.OutputConnector.set_parallelization`
                                method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :meth:`~connectors.connectors.OutputConnector.set_executor`
                         method for details
        """
        Connector.__init__(self, instance, method, parallelization, executor)
        self.__caching = caching
        self.__announcements = weakref.WeakSet()
        self.__connections = set()          # stores tuples (connector, instance). The instance is only saved to prevent its deletion through reference counting
        self.__result = None
        self.__result_is_valid = False
        self.__observed_has_changed = True
        self.__running = False              # is used to prevent, that the getter is executed multiple times for the same changes
        self.__computable = common.Event()  # is set, when there is no pending announcement
        self.__computable.set()

    def __call__(self, *args, **kwargs):
        """By making the object callable, it mimics the replaced method.
        This method also notifies the input connectors, that are connected to
        this output.

        :param `*args,**kwargs`: parameters with which the replaced method has been called
        :returns: the return value of the replaced method
        """
        if self.__result_is_valid:
            return self.__result
        return self._executor.run_coroutine(self._request(self._executor, *args, **kwargs))

    def connect(self, connector):
        """A method for connecting this output connector to an input connector.

        :param connector: the input connector to which this connector shall be connected
        :returns: the instance of which this :class:`OutputConnector` has replaced a method
        """
        for c in connector._connect(self):
            self.__connections.add((c, c._get_instance()))
        return self._instance()

    def disconnect(self, connector):
        """A method for disconnecting this output connector from an input connector,
        to which it is currently connected.

        :param connector: the input connector from which this connector shall be disconnected
        :returns: the instance of which this :class:`OutputConnector` has replaced a method
        """
        for c in connector._disconnect(self):
            self.__connections.remove((c, c._get_instance()))
        return self._instance()

    def set_caching(self, caching):
        """Specifies, if the result value of this output connector shall be cached.
        If caching is enabled and the result value is retrieved (e.g. through a
        connection or by calling the connector), the cached value is returned and
        the replaced getter method is not called unless the result value has to
        be re-computed, because an observed setter method has changed a parameter
        for the computation. In this case, the getter method is only called once,
        independent of the number of connections through which the result value
        has to be passed.

        :param caching: True, if caching shall be enabled, False otherwise
        """
        self.__caching = caching
        if not self.__caching:
            self.__result_is_valid = False
            self.__result = None

    def _announce(self, connector, non_lazy_inputs):
        """This method is to notify this output connector, when an observed input
        connector (a setter from self._instance) can retrieve updated data.

        :param connector: the input connector, which is about to change a value
        :param non_lazy_inputs: a :class:`~connectors._common._non_lazy_inputs.NonLazyInputs`
                                instance to which input connectors can be appended,
                                if they request an immediate re-computation (see
                                the :class:`~connectors._connectors._base_classes.InputConnector`'s
                                :meth:`~connectors._connectors._base_classes.InputConnector.set_laziness`
                                method for more about lazy execution)
        """
        self.__announcements.add(connector)
        self.__result_is_valid = False
        self.__computable.clear()
        for c, _ in self.__connections:
            c._announce(self, non_lazy_inputs)

    def _notify(self, connector):
        """This method is to notify this output connector, when an observed input
        connector (a setter from self._instance) has retrieved updated data.

        :param connector: the input connector, which has changed a value
        :param non_lazy_inputs: a :class:`~connectors._common._non_lazy_inputs.NonLazyInputs`
                                instance to which input connectors can be appended,
                                if they request an immediate re-computation (see
                                the :class:`~connectors._connectors._base_classes.InputConnector`'s
                                :meth:`~connectors._connectors._base_classes.InputConnector.set_laziness`
                                method for more about lazy execution)
        """
        self.__result_is_valid = False
        self.__observed_has_changed = True
        self.__result = None
        self.__announcements.discard(connector)
        if not self.__announcements:
            self.__computable.set()

    def _cancel(self, connector):
        """Notifies this output connector, that an announced value change is not
        going to happen.

        :param connector: the observed input connector whose value change is canceled
        """
        self.__announcements.discard(connector)
        if not self.__announcements:
            self.__computable.set()
            self.__result_is_valid = not self.__observed_has_changed    # if all announcements have been canceled, the cached result is still valid
            for c, _ in self.__connections:
                c._cancel(self)     # pylint: disable=protected-access; the _cancel method is meant to be called from other connectors, but not from outside this package

    async def _request(self, executor, *args, **kwargs):
        """Causes this output connector to re-compute its value and notifies the
        connected input connectors.
        This method is called by a connected input connector, when it needs the
        result value of this output connector.

        :param connector: the input connector, from which the request is issued
        :param executor: the :class:`~connectors._common._executors.Executor` instance,
                         that manages the current computations
        :returns: the result value of the output connector
        """
        if not self.__running:
            self.__running = True
            try:
                if self.__result_is_valid:
                    if self.__connections:
                        await asyncio.wait([c._notify(self, self.__result, executor) for c, _ in self.__connections])
                    return self.__result
                else:
                    # wait for the announced value changes
                    if self.__announcements:
                        await asyncio.wait([a._request(executor) for a in self.__announcements])
                        await self.__computable.wait(executor)
                        if self.__result_is_valid:  # this can happen, if all announcements have been canceled
                            return self.__result
                    # execute the getter
                    result = await executor.run_method(self._parallelization, self._method,
                                                       self._instance(), *args, **kwargs)
                    if self.__caching:
                        self.__result = result
                        self.__result_is_valid = True
                        self.__observed_has_changed = False
                    # notify the connected inputs
                    if self.__connections:
                        await asyncio.wait([c._notify(self, result, executor) for c, _ in self.__connections])
                    return result
            finally:
                self.__running = False



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains the :class:`~connectors.MultiInputData` class"""

import collections

__all__ = ("MultiInputData",)


class MultiInputData(collections.OrderedDict):
    # pylint: disable=wrong-spelling-in-docstring;    avoid that the spell checker complains about the code in this comment
    """A container for data that is managed with a multi-input connector.
    This is basically an :class:`~collections.OrderedDict` with an add method,
    that stores the added data under a unique key.
    This facilitates the implementation of a class with a :class:`~connectors.MultiInput` connector:

    >>> import connectors
    >>> class ReplacingMultiInput:
    ...     def __init__(self):
    ...         self.__data = connectors.MultiInputData()
    ...
    ...     @connectors.MultiInput()
    ...     def add_value(self, value):
    ...         return self.__data.add(value)
    ...
    ...     @add_value.remove
    ...     def remove_value(self, data_id):
    ...         del self.__data[data_id]
    ...
    ...     @add_value.replace
    ...     def replace_value(self, data_id, value):
    ...         self.__data[data_id] = value

    """

    def __init__(self, datas=()):
        """
        :param datas: an optional sequence of data objects, that shall be added to the container
        """
        collections.OrderedDict.__init__(self)
        self.__last_id = 0
        for data in datas:
            self.add(data)

    def add(self, data):
        """Adds a data set to the container.

        :param data: the data set that shall be added
        :returns: the id under which the data is stored
        """
        while self.__last_id in self:
            self.__last_id += 1
        self[self.__last_id] = data
        return self.__last_id



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains the :class:`~connectors._common._executors.Executor` classes, that
implement the parallelization of computations in separate threads or processes.

Only the factory function for creating :class:`~connectors._common._executors.Executor`
instances is exported.
"""

# pylint: disable=wrong-spelling-in-comment,wrong-spelling-in-docstring;    for some reason the spell checker does not recognize the word "CPU"

import asyncio
import concurrent.futures
import os
from connectors.connectors import Connector
from connectors._common import Parallelization

__all__ = ("executor",)


def executor(threads=None, processes=0):
    """A factory function for creating :class:`~connectors._common._executors.Executor`
    objects. Executors define how the computations of a processing chain are
    parallelized by executing them in separate threads or processes. This function
    creates an executor and configures it to use at maximum the given number of
    threads or processes.

    :param threads: an integer number of threads or ``None`` to determine the number
                    automatically. 0 disables the thread based parallelization.
    :param processes: an integer number of processes or ``None`` to determine the
                      number automatically (in this case, the number of CPU cores
                      will be taken). 0 disables the process based parallelization.
    """
    if threads == 0:
        if processes == 0:
            return SequentialExecutor()
        else:
            return MultiprocessingExecutor(number_of_processes=processes)
    else:
        if processes == 0:
            return ThreadingExecutor(number_of_threads=threads)
        else:
            return ThreadingMultiprocessingExecutor(number_of_threads=threads, number_of_processes=processes)


def _redeployed_method(method_name, reduced_instance, *args, **kwargs):
    """A helper function, that is passed to the separate processes for executing
    the given method.
    This function is necessary, because objects with connectors cannot be pickled,
    so that their methods cannot be passed to a process directly. Instead only the
    relevant data is serialized and passed to the process, where this function
    unwraps this data and executes the method.
    :param method_name: the string name of the method that shall be executed
    :param reduced_instance: a tuple with the class, in which the method is defined
                             and the ``__dict__`` of the instance of which the
                             method shall be executed
    :param `*args,**kwargs`: arguments for the method
    """
    class_, state = reduced_instance
    instance = class_.__new__(class_)
    instance.__dict__.update(state)
    method = getattr(instance, method_name)
    return method(*args, **kwargs)


class Executor:
    """a base class for managing the event loop and the execution in threads or processes."""

    def __init__(self):
        self._loop = None   # will be initialized in run_coroutine or run_until_complete
        self.__tasks = []

    async def run_method(self, parallelization, method, instance, *args, **kwargs):
        """Abstract method, whose overrides shall execute the given method.
        The parallelization shall be implemented in this method.

        :param parallelization: a flag of :class:`connectors.Parallelization`, that
                                specifies how the given method can be parallelized
        :param method: the unbound method, that shall be executed
        :param instance: the instance of which the method shall be executed
        :param `*args,**kwargs`: arguments for the method
        :returns: the return value of the method
        """
        raise NotImplementedError("this method should have been overridden in a derived class")

    def run_coroutine(self, coro):
        """Takes a coroutine and runs it in a newly created event loop.

        :param coro: the coroutine
        :returns: the return value of the coroutine
        """
        self._set_up()
        try:
            task = self._loop.create_task(coro)
            return self._loop.run_until_complete(task)
        finally:
            self._tear_down()

    def run_until_complete(self, future):
        """Takes a future or a task and runs it in a newly created event loop.
        This is a wrapper for the event loop's :meth:`~asyncio.loop.run_until_complete`
        method.

        :param future: the future or the task
        :returns: the return value of the execution
        """
        self._set_up()
        try:
            return self._loop.run_until_complete(future)
        finally:
            self._tear_down()

    def get_event_loop(self):
        """Returns the currently active event loop. This can be None, if no
        coroutine, task or future is currently being processed.

        :returns: the event loop or None
        """
        return self._loop

    def _set_up(self):
        """Is called by the run_until_complete and run_coroutine methods before
        executing the passed object.

        This implementation creates the event loop. This method can be overwritten
        in order to instantiate other objects such as thread pools, which are
        required during the execution of the passed object. In this case, make
        sure to call the overridden base class's method to create the event loop
        aswell.
        """
        self._loop = asyncio.new_event_loop()

    def _tear_down(self):
        """Is called by the run_until_complete and run_coroutine methods after
        executing the passed object.

        This implementation closes the event loop. This method can be overwritten
        in order to tear down other objects such as thread pools, which were
        required during the execution of the passed object. In this case, make
        sure to call the overridden base class's method to close the event loop
        aswell.
        """
        self._loop.close()
        self._loop = None


class SequentialExecutor(Executor):
    """An executor class, that executes everything sequentially."""

    async def run_method(self, parallelization, method, instance, *args, **kwargs):
        """Executes the given method sequentially.

        :param parallelization: a flag of :class:`connectors.Parallelization`, that
                                specifies how the given method can be parallelized
        :param method: the unbound method, that shall be executed
        :param instance: the instance of which the method shall be executed
        :param `*args,**kwargs`: arguments for the method
        :returns: the return value of the method
        """
        return method(instance, *args, **kwargs)


class ThreadingExecutor(Executor):
    """An executor class, that can parallelize computations with threads."""

    def __init__(self, number_of_threads):
        """
        :param number_of_threads: the maximum number of threads, that shall be
                                  created, or None to determine this number
                                  automatically.
        """
        Executor.__init__(self)
        self.__number_of_threads = number_of_threads
        self.__executor = None  # will be initialized in run_coroutine or run_until_complete

    async def run_method(self, parallelization, method, instance, *args, **kwargs):
        """Executes the given method in a thread if possible and falls back to
        sequential execution if not.

        :param parallelization: a flag of :class:`connectors.Parallelization`, that
                                specifies how the given method can be parallelized
        :param method: the unbound method, that shall be executed
        :param instance: the instance of which the method shall be executed
        :param `*args,**kwargs`: arguments for the method
        :returns: the return value of the method
        """
        if parallelization == Parallelization.SEQUENTIAL:
            return method(instance, *args, **kwargs)
        else:
            return await self._loop.run_in_executor(self.__executor, method, instance, *args, **kwargs)

    def _set_up(self):
        """Is called by the run_until_complete and run_coroutine methods before
        executing the passed object.

        This implementation instantiates the ThreadPoolExecutor and calls the
        overridden method to create the event loop.
        """
        super()._set_up()
        if self.__number_of_threads is None:            # the default number of workers for the ThreadPoolExecutor is 5x the CPU count, which is meant for I/O bound tasks.
            self.__number_of_threads = os.cpu_count()   # This class is meant for CPU work, so the number of threads should be lower to reduce context switching overhead.
        self.__executor = concurrent.futures.ThreadPoolExecutor(max_workers=self.__number_of_threads)

    def _tear_down(self):
        """Is called by the run_until_complete and run_coroutine methods after
        executing the passed object.

        This implementation shuts down the ThreadPoolExecutor and calls the
        overridden method to close the event loop.
        """
        self.__executor.shutdown()
        super()._tear_down()


class MultiprocessingExecutor(Executor):
    """An executor class, that can parallelize computations with processes."""

    def __init__(self, number_of_processes):
        """
        :param number_of_processes: the maximum number of processes, that shall be
                                    created, or None to determine this number
                                    automatically (in this case, the number of CPU
                                    cores will be taken).
        """
        Executor.__init__(self)
        self.__number_of_processes = number_of_processes
        self.__executor = None  # will be initialized in run_coroutine or run_until_complete

    async def run_method(self, parallelization, method, instance, *args, **kwargs):
        """Executes the given method in a process if possible and falls back to
        sequential execution if not.

        :param parallelization: a flag of :class:`connectors.Parallelization`, that
                                specifies how the given method can be parallelized
        :param method: the unbound method, that shall be executed
        :param instance: the instance of which the method shall be executed
        :param `*args,**kwargs`: arguments for the method
        :returns: the return value of the method
        """
        if parallelization == Parallelization.PROCESS:
            class_ = instance.__class__
            state = instance.__dict__.copy()
            to_remove = []
            for a in state:
                if isinstance(state[a], Connector):
                    to_remove.append(a)
            for a in to_remove:
                del state[a]
            reduced_instance = (class_, state)
            return await self._loop.run_in_executor(self.__executor,
                                                    _redeployed_method,
                                                    method.__name__,
                                                    reduced_instance,
                                                    *args, **kwargs)
        else:
            return method(instance, *args, **kwargs)

    def _set_up(self):
        """Is called by the run_until_complete and run_coroutine methods before
        executing the passed object.

        This implementation instantiates the ProcessPoolExecutor and calls the
        overridden method to create the event loop.
        """
        super()._set_up()
        self.__executor = concurrent.futures.ProcessPoolExecutor(max_workers=self.__number_of_processes)

    def _tear_down(self):
        """Is called by the run_until_complete and run_coroutine methods after
        executing the passed object.

        This implementation shuts down the ProcessPoolExecutor and calls the
        overridden method to close the event loop.
        """
        self.__executor.shutdown()
        super()._tear_down()


class ThreadingMultiprocessingExecutor(Executor):
    """An executor class, that can parallelize computations with both threads and processes."""

    def __init__(self, number_of_threads, number_of_processes):
        """
        :param number_of_threads: the maximum number of threads, that shall be
                                  created, or None to determine this number
                                  automatically.
        :param number_of_processes: the maximum number of processes, that shall be
                                    created, or None to determine this number
                                    automatically (in this case, the number of CPU
                                    cores will be taken).
        """
        Executor.__init__(self)
        self.__number_of_threads = number_of_threads
        self.__number_of_processes = number_of_processes
        self.__thread_executor = None   # will be initialized in run_coroutine or run_until_complete
        self.__process_executor = None  # will be initialized in run_coroutine or run_until_complete

    async def run_method(self, parallelization, method, instance, *args, **kwargs):
        """Executes the given method in a process if possible and falls back to
        threaded and then sequential execution if not.

        :param parallelization: a flag of :class:`connectors.Parallelization`, that
                                specifies how the given method can be parallelized
        :param method: the unbound method, that shall be executed
        :param instance: the instance of which the method shall be executed
        :param `*args,**kwargs`: arguments for the method
        :returns: the return value of the method
        """
        if parallelization == Parallelization.SEQUENTIAL:
            return method(instance, *args, **kwargs)
        elif parallelization == Parallelization.THREAD:
            return await self._loop.run_in_executor(self.__thread_executor, method, instance, *args, **kwargs)
        else:
            class_ = instance.__class__
            state = instance.__dict__.copy()
            to_remove = []
            for a in state:
                if isinstance(state[a], Connector):
                    to_remove.append(a)
            for a in to_remove:
                del state[a]
            reduced_instance = (class_, state)
            return await self._loop.run_in_executor(self.__process_executor,
                                                    _redeployed_method,
                                                    method.__name__,
                                                    reduced_instance,
                                                    *args, **kwargs)

    def _set_up(self):
        """Is called by the run_until_complete and run_coroutine methods before
        executing the passed object.

        This implementation instantiates the ThreadPoolExecutor and the
        ProcessPoolExecutor and calls the overridden method to create the event
        loop.
        """
        super()._set_up()
        if self.__number_of_threads is None:            # the default number of workers for the ThreadPoolExecutor is 5x the CPU count, which is meant for I/O bound tasks.
            self.__number_of_threads = os.cpu_count()   # This class is meant for CPU work, so the number of threads should be lower to reduce context switching overhead.
        self.__thread_executor = concurrent.futures.ThreadPoolExecutor(max_workers=self.__number_of_threads)
        self.__process_executor = concurrent.futures.ProcessPoolExecutor(max_workers=self.__number_of_processes)

    def _tear_down(self):
        """Is called by the run_until_complete and run_coroutine methods after
        executing the passed object.

        This implementation shuts down the ThreadPoolExecutor and the
        ProcessPoolExecutorand calls the overridden method to close the event
        loop.
        """
        self.__thread_executor.shutdown()
        self.__process_executor.shutdown()
        super()._tear_down()



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains the :class:`MultiInput` class for decorating special setter methods, with
which multiple values can be added.
Also contains a descriptor class for the remove and replace methods of a multi-input.
"""

from .. import _common as common
from .._proxies import MultiInputProxy
from ._baseclasses import InputDecorator, default_executor

__all__ = ("MultiInput",)


class MultiInput(InputDecorator):
    """A decorator, that marks a method as an input for multiple connections.
    These connections can be used to automatically update a processing chain
    when a value has changed.

    The decorated method must take exactly one argument and return a unique id.
    For every MultiInput-method a remove method has to be provided, which removes
    a value, that has previously been added.

    A replace method can be provided optionally. This method is called, when the
    value of a connected output has changed, rather than removing the old value and
    adding the new.

    See the :meth:`~connectors.MultiInput.remove` and :meth:`~connectors.MultiInput.replace`
    methods for documentation about how to define these methods for a multi-input connector.
    """

    def __init__(self,
                 observers=(),
                 laziness=common.Laziness.get_default(),
                 parallelization=common.Parallelization.get_default_multiinput_parallelization(),
                 executor=default_executor):
        """
        :param observers: the names of output methods that are affected by passing
                          a value to this connector. For convenience it is also
                          possible to pass a string here, if only one output
                          connector depends on this input.
        :param laziness: a flag from the :class:`connectors.Laziness` enum. See the
                         :class:`~connectors.connectors.MultiInputConnector`'s
                         :meth:`~connectors.connectors.MultiInputConnector.set_laziness`
                         method for details
        :param parallelization: a flag from the :class:`connectors.Parallelization` enum.
                                See the :class:`~connectors.connectors.MultiInputConnector`'s
                                :meth:`~connectors.connectors.MultiInputConnector.set_parallelization`
                                method for details
        :param executor: an :class:`~connectors._common._executors.Executor` instance,
                         that can be created with the :func:`connectors.executor`
                         function. See the :class:`~connectors.connectors.MultiInputConnector`'s
                         :meth:`~connectors.connectors.MultiInputConnector.set_executor`
                         method for details
        """
        InputDecorator.__init__(self,
                                observers=observers,
                                laziness=laziness,
                                parallelization=parallelization,
                                executor=executor)
        self.__remove_method = None
        self.__replace_method = None

    def __get__(self, instance, instance_type):
        """Is called, when the decorated method is accessed.

        :param instance: the instance of which a method shall be replaced
        :param instance_type: the type of the instance
        :returns: a :class:`MultiInputProxy` instance, that mimics the decorated
                  method and adds the connector functionality
        """
        return MultiInputProxy(instance=instance,
                               method=self._method,
                               remove_method=self.__remove_method,
                               replace_method=self.__replace_method,
                               observers=self._observers,
                               announce_condition=self._announce_condition,
                               notify_condition=self._notify_condition,
                               laziness=self._laziness,
                               parallelization=self._parallelization,
                               executor=self._executor)

    def remove(self, method):
        """A method of the decorated method to decorate the remove method, with
        which data, that has been added through the decorated method can be removed.

        A remove method has to take the ID, which has been returned by the multi-input
        method as a parameter, so it knows which value has to be removed.

        The usage is described by the following example: if `A` is the name of
        the method, that is decorated with :class:`MultiInput`, the remove method
        has to be decorated with ``@A.remove``.

        :param method: the decorated remove method
        :returns: a MultiInputAssociateDescriptor, that generates a MultiInputAssociateProxy,
                  which enhances the decorated method with the functionality, that
                  is required for the multi-input connector
        """
        self.__remove_method = method
        return common.MultiInputAssociateDescriptor(method=method,
                                                    observers=self._observers,
                                                    parallelization=self._parallelization,
                                                    executor=self._executor)

    def replace(self, method):
        """A method of the decorated method to decorate the replace method, with
        which data, that has been added through the decorated method, can be replaced
        with new data without changing the ID under which it is stored.

        A replace method has to take the ID, under which the old data is stored,
        as first parameter and the new data as second parameter. It is strongly
        recommended, that this method stores the new data under the same ID as
        the old data. And this method must return the ID, under which the new
        data is stored.

        Also, if no data has been stored under the given ID, yet, the replace method
        shall simply store the data under the given ID instead of raising an error.

        Specifying a replace method is optional. If no method has been decorated
        to be a replace method, the :class:`MultiInput` connector falls back to removing
        the old data and adding the updated one, whenever updated data is propagated
        through its connections.

        The usage is described by the following example: if `A` is the name of
        the method, that is decorated with :class:`MultiInput`, the remove method
        has to be decorated with ``@A.replace``.

        :param method: the decorated replace method
        :returns: a MultiInputAssociateDescriptor, that generates a MultiInputAssociateProxy,
                  which enhances the decorated method with the functionality, that
                  is required for the multi-input connector
        """
        self.__replace_method = method
        return common.MultiInputAssociateDescriptor(method=method,
                                                    observers=self._observers,
                                                    parallelization=self._parallelization,
                                                    executor=self._executor)



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Contains the :class:`Input` class for decorating setter methods"""

from .._proxies import SingleInputProxy
from ._baseclasses import InputDecorator

__all__ = ("Input",)


class Input(InputDecorator):
    """A decorator, that marks a method as an input for single connections.
    These connections can be used to automatically update a processing chain
    when a value has changed.
    The decorated method must take exactly one argument.
    """

    def __get__(self, instance, instance_type):
        """Is called, when the decorated method is accessed.

        :param instance: the instance of which a method shall be replaced
        :param instance_type: the type of the instance
        :returns: a :class:`~connectors.proxies.SingleInputProxy` instance, that mimics
                  the decorated method and adds the connector functionality
        """
        return SingleInputProxy(instance=instance,
                                method=self._method,
                                observers=self._observers,
                                announce_condition=self._announce_condition,
                                notify_condition=self._notify_condition,
                                laziness=self._laziness,
                                parallelization=self._parallelization,
                                executor=self._executor)



# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Helper functions for input connectors"""

import inspect

__all__ = ("select_condition_methods", "resolve_observers", "forward_announcement", "get_first_argument")


def non_condition(*args, **kwargs):   # pylint: disable=unused-argument; this method has to be compatible with multiple input connector implementations
    """A condition function for the announce- or notify condition of an input
    connector, if that condition is not specified.

    :returns: True
    """
    return True


def select_condition_methods(announce_condition, notify_condition):
    """Selects the announce- and notify conditions, in case any of them is None

    :param announce_condition: a method for the announce condition or None
    :param notify_condition: a method for the notify condition or None
    :returns: a tuple of the announce- and notify condition methods, neither of which is None
    """
    return (non_condition if announce_condition is None else announce_condition,
            non_condition if notify_condition is None else notify_condition)


def resolve_observers(instance, observers):
    """Gets the observing output connectors for an input connector

    :param instance: the instance, to which the input and the output connectors belong
    :param observers: a sequence of string method names
    :returns: a tuple of OutputConnector instances
    """
    return [getattr(instance, o)._get_connector() for o in observers]


def forward_announcement(connector, observers, non_lazy_inputs, laziness):
    """Forwards the announcement, that an input connector has received, to the
    observing output connectors.

    :param connector: the input connector, that has received the announcement
    :param observers: a sequence of observing output connectors, to which the
                      announcement shall be forwarded
    :param non_lazy_inputs: a :class:`~connectors._common._non_lazy_inputs.NonLazyInputs`
                            instance to which input connectors can be appended,
                            if they request an immediate re-computation (see the
                            :class:`~connectors._connectors._base_classes.InputConnector`'s
                            :meth:`~connectors._connectors._base_classes.InputConnector.set_laziness`
                            method for more about lazy execution)
    :param laziness: the laziness setting of the input connector
    """
    initial_number_of_non_lazy_inputs = len(non_lazy_inputs)
    for o in observers:
        o._announce(connector, non_lazy_inputs)
    if len(non_lazy_inputs) == initial_number_of_non_lazy_inputs:   # only add self, if no methods down the processing chain requests its execution anyway
        non_lazy_inputs.add(connector=connector, laziness=laziness)


def get_first_argument(method, *args, **kwargs):
    """Gets the first argument, that has been passed to a method, from the *args
    and **kwargs parameters.

    When using this function for an unbound method, don't pass the 'self' argument
    to this function, because otherwise, this would be the one, that is always
    returned.

    :param method: the method, of whose call, the first parameter shall be determined
    :param `*args,**kwargs`: the parameters, that have been passed to the method
    :returns: the first argument, that has been passed to the method, or None if no argument has been passed
    """
    if args:
        return args[0]
    elif len(kwargs) == 1:  # a shortcut to avoid introspection
        return next(iter(kwargs.values()))
    else:                   # get the first argument of the method, that is given in the **kwargs dict
        for a in inspect.signature(method).parameters:
            if a in kwargs:
                return kwargs[a]
    return None
